# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module contains functions for RETIS.

This module defines functions that are needed to perform Replica
Exchange Transition Interface Sampling (RETIS). The RETIS algorithm
was first described by van Erp [RETIS]_.


Methods defined here
~~~~~~~~~~~~~~~~~~~~

make_retis_step (:py:func:`.make_retis_step`)
    Function to select and execute the RETIS move.

retis_tis_moves (:py:func:`.retis_tis_moves`)
    Function to execute the TIS steps in the RETIS algorithm.

retis_moves (:py:func:`.retis_moves`)
    Function to perform RETIS swapping moves - it selects what scheme
    to use, i.e. ``[0^-] <-> [0^+], [1^+] <-> [2^+], ...`` or
    ``[0^+] <-> [1^+], [2^+] <-> [3^+], ...``.

retis_swap_wrapper (:py:func:`.retis_swap_wrapper`)
    The function that actually swaps two path ensembles.
    This function decides which swap will be done:
    it can be a typical RETIS swap (retis_swap), or a swap between two PPTIS
    ensembles with limited memory (prretis_swap).

retis_swap (:py:func:`.retis_swap`)
    The function that actually swaps two path ensembles.
    A swap between two typical RETIS ensembles, not PPTIS ensembles.

repptis_swap (:py:func:`.repptis_swap`)
    The function that actually swaps two path ensembles.
    A swap between two PPTIS ensembles with truncated paths,
    so this is not the typical RETIS swap.

retis_swap_zero (:py:func:`.retis_swap_zero`)
    The function that performs the swapping for the
    ``[0^-] <-> [0^+]`` swap.

high_acc_swap (:py:func:`.high_acc_wap`)
    The function coputes if a path generated via SS can be accepted
    for swapping in accordance to super detail balance.

References
~~~~~~~~~~
.. [RETIS] T. S. van Erp,
   Phys. Rev. Lett. 98, 26830 (2007),
   http://dx.doi.org/10.1103/PhysRevLett.98.268301

"""
import copy
import logging
from pyretis.core.common import (null_move,
                                 compute_weight,
                                 priority_checker)
from pyretis.core.tis import make_tis, paste_paths
from pyretis.inout.restart import write_ensemble_restart

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

# pylint: disable=W0106

__all__ = ['high_acc_swap',
           'make_retis_step',
           'repptis_swap',
           'retis_moves',
           'retis_swap_wrapper',
           'retis_swap',
           'retis_swap_zero']


def make_retis_step(ensembles, rgen, settings, cycle):
    """Determine and execute the appropriate RETIS move.

    Here we will determine what kind of RETIS moves we should do.
    We have two options:

    1) Do the RETIS swapping moves. This is done by calling
       :py:func:`.retis_moves`.
    2) Do TIS moves, either for all ensembles or for just one, based on
       values of relative shoot frequencies. This is done by calling
       :py:func:`.retis_tis_moves`.

    This function will just determine and execute the appropriate move
    (1 or 2) based on the given swapping frequencies in the `settings`
    and drawing a random number from the random number generator `rgen`.

    Parameters
    ----------
    ensembles : list of dicts of objects
        This is a list of the ensembles we are using in the RETIS method.
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    rgen : object like :py:class:`.RandomGenerator`
        This is a random generator. Here we assume that we can call
        `rgen.rand()` to draw random uniform numbers.
    settings : dict
        Contains the settings for the simulation.
    cycle : integer
        The current cycle number.

    Returns
    -------
    out : generator
        This generator yields the results after performing the
        RETIS moves.

    """
    prio_skip = priority_checker(ensembles, settings)
    swap_freq = settings['retis']['swapfreq']
    swap = False if True in prio_skip else rgen.rand() < swap_freq

    if swap:
        # Do RETIS moves
        logger.info('Performing RETIS swapping move(s).')
        results = retis_moves(ensembles, rgen, settings, cycle)
    else:
        logger.info('Performing RETIS TIS move(s)')
        results = make_tis(ensembles, rgen, settings, cycle)
    return results


def retis_moves(ensembles, rgen, settings, cycle):
    """Perform RETIS moves on the given ensembles.

    This function will perform RETIS moves on the given ensembles.
    First we have two strategies based on
    `settings['retis']['swapsimul']`:

    1) If `settings['retis']['swapsimul']` is True we will perform
       several swaps, either ``[0^-] <-> [0^+], [1^+] <-> [2^+], ...``
       or ``[0^+] <-> [1^+], [2^+] <-> [3^+], ...``. Which one of these
       two swap options we use is determined randomly and they have
       equal probability.

    2) If `settings['retis']['swapsimul']` is False we will just
       perform one swap for randomly chosen ensembles, i.e. we pick a
       random ensemble and try to swap with the ensemble to the right.
       Here we may also perform null moves if the
       `settings['retis']['nullmove']` specifies so.

    Parameters
    ----------
    ensembles : list of dicts of objects
        Each contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    rgen : object like :py:class:`.RandomGenerator`
        This is a random generator. Here we assume that we can call
        `rgen.rand()` to draw random uniform numbers.
    settings : dict
        This dict contains the settings for the RETIS method.
    cycle : integer
        The current cycle number.

    Yields
    ------
    out : dict
        This dictionary contains the result of the RETIS moves.

    """
    if settings['retis']['swapsimul']:
        # Here we have two schemes:
        # 1) scheme == 0: [0^-] <-> [0^+], [1^+] <-> [2^+], ...
        # 2) scheme == 1: [0^+] <-> [1^+], [2^+] <-> [3^+], ...
        if len(ensembles) < 3:
            # Small number of ensembles, can only do the [0^-] <-> [0^+] swap:
            scheme = 0
        else:
            scheme = 0 if rgen.rand() < 0.5 else 1
        for i in range(scheme, len(ensembles) - 1, 2):
            idx = ensembles[i]['path_ensemble'].ensemble_number
            accept, trial, status = retis_swap_wrapper(
                ensembles, idx, settings, cycle
            )
            result = {
                'ensemble_number': idx,
                'mc-move': 'swap',
                'status': status,
                'trial': trial[0],
                'accept': accept,
                'swap-with': idx + 1,
            }
            yield result
            result = {
                'ensemble_number': idx + 1,
                'mc-move': 'swap',
                'status': status,
                'trial': trial[1],
                'accept': accept,
                'swap-with': idx,
            }
            yield result
        # We might have missed some ensembles in the two schemes.
        # Here, we do null moves in these, if requested:
        if settings['retis']['nullmoves']:
            if len(ensembles) % 2 != scheme:  # Missed last ensemble:
                # This is perhaps strange but it is equivalent to:
                # (scheme == 0 and len(ensembles) % 2 != 0) or
                # (scheme == 1 and len(ensembles) % 2 == 0)
                accept, trial, status = null_move(ensembles[-1], cycle)
                result = {
                    'ensemble_number':
                        ensembles[-1]['path_ensemble'].ensemble_number,
                    'mc-move': 'nullmove',
                    'status': status,
                    'trial': trial,
                    'accept': accept,
                }
                yield result
            # We always miss the first ensemble in scheme 1:
            if scheme == 1:
                accept, trial, status = null_move(ensembles[0], cycle)
                result = {
                    'ensemble_number':
                        ensembles[0]['path_ensemble'].ensemble_number,
                    'mc-move': 'nullmove',
                    'status': status,
                    'trial': trial,
                    'accept': accept,
                }
                yield result
    else:  # Just swap two ensembles:
        idx = rgen.random_integers(0, len(ensembles) - 2)
        accept, trial, status = retis_swap_wrapper(
            ensembles, idx, settings, cycle
        )
        result = {
            'ensemble_number': idx,
            'mc-move': 'swap',
            'status': status,
            'trial': trial[0],
            'accept': accept,
            'swap-with': idx + 1,
        }
        yield result
        result = {
            'ensemble_number': idx + 1,
            'mc-move': 'swap',
            'status': status,
            'trial': trial[1],
            'accept': accept,
            'swap-with': idx,
        }
        yield result
        # Do null moves in the other ensembles, if requested:
        if settings['retis']['nullmoves']:
            for ensemble in ensembles:
                idx2 = ensemble['path_ensemble'].ensemble_number
                if idx2 not in (idx, idx + 1):
                    accept, trial, status = null_move(ensemble, cycle)
                    result = {
                        'ensemble_number': idx2,
                        'mc-move': 'nullmove',
                        'status': status,
                        'trial': trial,
                        'accept': accept,
                    }
                    yield result


def retis_swap_wrapper(ensembles, idx, settings, cycle):
    """Swap the last accepted paths of two TIS or two PPTIS ensembles.

    This function selects the correct swap function to use, based on the
    simulation task (retis or repptis).

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the RE(PP)TIS method.
        Each one contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
            This is used for storing results for the simulation. It
            is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
            System is used here since we need access to the temperature
            and to the particle list
        * `order_function`: object like :py:class:`.OrderParameter`
            The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
            The engine to use for propagating a path.

    idx : integer
        Definition of what path ensembles to swap. We will swap
        `ensembles[idx]` with `ensembles[idx+1]`.
    settings : dict
        This dict contains the settings for the RE(PP)TIS method.
    cycle : integer
        Current cycle number.

    Returns
    -------
    out[0] : boolean
        Should the path be accepted or not?
    out[1] : list of object like :py:class:`.PathBase`
        The trial paths.
    out[2] : string
        The status for the trial paths.

    """
    if settings['simulation']['task'] == "repptis":
        return repptis_swap(ensembles, idx, settings, cycle)

    return retis_swap(ensembles, idx, settings, cycle)


def retis_swap(ensembles, idx, settings, cycle):
    """Perform a RETIS swapping move for two ensembles.

    These ensembles are not PPTIS ensembles.

    The RETIS swapping move will attempt to swap accepted paths between
    two ensembles in the hope that path from [i^+] is an acceptable path
    for [(i+1)^+] as well. We have two cases:

    1) If we try to swap between [0^-] and [0^+] we need to integrate
       the equations of motion.
    2) Otherwise, we can just swap and accept if the path from [i^+] is
       an acceptable path for [(i+1)^+]. The path from [(i+1)^+] is
       always acceptable for [i^+] (by construction).

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the RETIS method.
        Each one contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    idx : integer
        Definition of what path ensembles to swap. We will swap
        `ensembles[idx]` with `ensembles[idx+1]`. If `idx == 0` we have
        case 1) defined above.
    settings : dict
        This dict contains the settings for the RETIS method.
    cycle : integer
        Current cycle number.

    Returns
    -------
    out[0] : boolean
        Should the path be accepted or not?
    out[1] : list of object like :py:class:`.PathBase`
        The trial paths.
    out[2] : string
        The status for the trial paths.

    Note
    ----
    Note that path.generated is **NOT** updated here. This is because
    we are just swapping references and not the paths. In case the
    swap is rejected updating this would invalidate the last accepted
    path.

    """
    logger.info("Swapping: %s <-> %s",
                ensembles[idx]['path_ensemble'].ensemble_name,
                ensembles[idx+1]['path_ensemble'].ensemble_name)

    if idx == 0:
        return retis_swap_zero(ensembles, settings, cycle)
    path_ensemble1 = ensembles[idx]['path_ensemble']
    path_ensemble2 = ensembles[idx+1]['path_ensemble']

    path1 = path_ensemble1.last_path
    path2 = path_ensemble2.last_path
    ens_moves = [settings['ensemble'][i]['tis'].get('shooting_move', 'sh')
                 for i in [idx, idx+1]]
    intf_w = [list(i) for i in (path_ensemble1.interfaces,
                                path_ensemble2.interfaces)]
    for i, j in enumerate([settings['ensemble'][k] for k in (idx, idx+1)]):
        if ens_moves[i] == 'wf':
            intf_w[i][2] = j['tis'].get('interface_cap', intf_w[i][2])

    # Check if path1 can be accepted in ensemble 2:
    cross = path1.check_interfaces(path_ensemble2.interfaces)[-1]
    accept = cross[1]
    status = 'ACC' if accept else 'NCR'

    if accept and settings['tis'].get('high_accept', False):
        if 'ss' in ens_moves or 'wf' in ens_moves:
            accept, status = high_acc_swap([path1, path2],
                                           ensembles[idx]['rgen'],
                                           intf_w[0], intf_w[1],
                                           ens_moves)
    if accept:  # Accept the swap:
        logger.info('Swap was accepted.')
        # To avoid overwriting files, we move the paths to the
        # generate directory here.
        path_ensemble2.move_path_to_generate(path1)
        path_ensemble1.move_path_to_generate(path2)
        for i, path, path_ensemble, flag in ((0, path2, path_ensemble1, 's+'),
                                             (1, path1, path_ensemble2, 's-')):
            path.status = status
            ens_set = settings['ensemble'][idx + i]
            move = ens_moves[i]
            path.weight = compute_weight(path, intf_w[i], move)\
                if (ens_set['tis'].get('high_accept', False) and
                    move in ('wf', 'ss')) else 1

            path.set_move(flag) if path.get_move() != 'ld' else \
                path.set_move('ld')

            # Then moved into accepted directory by the `add_path_data` below.
            path_ensemble.add_path_data(path, status, cycle=cycle)
            if cycle % ens_set.get('output', {}).get('restart-file', 1) == 0:
                write_ensemble_restart(ensembles[idx+i], ens_set)

        return accept, (path2, path1), status

    logger.info('Swap was rejected. (%s)', status)

    # Make shallow copies:
    trial1 = copy.copy(path2)
    trial2 = copy.copy(path1)
    trial1.set_move('s+')  # Came from right.
    trial2.set_move('s-')  # Came from left.
    # Calculate weights:
    for i, trial in ((0, trial1), (1, trial2)):
        trial.weight = compute_weight(trial, intf_w[i], ens_moves[i])\
            if (settings['ensemble'][idx+i]['tis'].get('high_accept', False)
                and ens_moves[i] in ('wf', 'ss')) else 1.
    path_ensemble1.add_path_data(trial1, status, cycle=cycle)
    path_ensemble2.add_path_data(trial2, status, cycle=cycle)

    return accept, (trial1, trial2), status


def retis_swap_zero(ensembles, settings, cycle):
    """Perform the RETIS swapping for ``[0^-] <-> [0^+]`` swaps.

    The RETIS swapping move for ensembles [0^-] and [0^+] requires some
    extra integration. Here we are generating new paths for [0^-] and
    [0^+] in the following way:

    1) For [0^-] we take the initial point in [0^+] and integrate
       backward in time. This is merged with the second point in [0^+]
       to give the final path. The initial point in [0^+] starts to the
       left of the interface and the second point is on the right
       side - i.e. the path will cross the interface at the end points.
       If we let the last point in [0^+] be called ``A_0`` and the
       second last point ``B``, and we let ``A_1, A_2, ...`` be the
       points on the backward trajectory generated from ``A_0`` then
       the final path will be made up of the points
       ``[..., A_2, A_1, A_0, B]``. Here, ``B`` will be on the right
       side of the interface and the first point of the path will also
       be on the right side.

    2) For [0^+] we take the last point of [0^-] and use that as an
       initial point to generate a new trajectory for [0^+] by
       integration forward in time. We also include the second last
       point of the [0^-] trajectory which is on the left side of the
       interface. We let the second last point be ``B`` (this is on the
       left side of the interface), the last point ``A_0`` and the
       points generated from ``A_0`` we denote by ``A_1, A_2, ...``.
       Then the resulting path will be ``[B, A_0, A_1, A_2, ...]``.
       Here, ``B`` will be on the left side of the interface and the
       last point of the path will also be on the left side of the
       interface.

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the RETIS method.
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    settings : dict
        This dict contains the settings for the RETIS method.
    cycle : integer
        The current cycle number.

    Returns
    -------
    out : string
        The result of the swapping move.

    """
    path_ensemble0 = ensembles[0]['path_ensemble']
    path_ensemble1 = ensembles[1]['path_ensemble']
    engine0, engine1 = ensembles[0]['engine'], ensembles[1]['engine']
    maxlen0 = settings['ensemble'][0]['tis']['maxlength']
    maxlen1 = settings['ensemble'][1]['tis']['maxlength']

    ens_moves = [settings['ensemble'][i]['tis'].get('shooting_move', 'sh')
                 for i in [0, 1]]
    intf_w = [list(i) for i in (path_ensemble0.interfaces,
                                path_ensemble1.interfaces)]
    for i, j in enumerate([settings['ensemble'][k] for k in (0, 1)]):
        if ens_moves[i] == 'wf':
            intf_w[i][2] = j['tis'].get('interface_cap', intf_w[i][2])

    # 0. check if MD is allowed
    allowed = (path_ensemble0.last_path.get_end_point(
                path_ensemble0.interfaces[0],
                path_ensemble0.interfaces[-1]) == 'R')
    # 1. Generate path for [0^-] from [0^+]:
    # We generate from the first point of the path in [0^+]:
    logger.debug('Swapping [0^-] <-> [0^+]')
    logger.debug('Creating path for [0^-]')
    system = path_ensemble1.last_path.phasepoints[0].copy()
    logger.debug('Initial point is: %s', system)
    ensembles[0]['system'] = system
    # Propagate it backward in time:
    path_tmp = path_ensemble1.last_path.empty_path(maxlen=maxlen1-1)
    if allowed:
        logger.debug('Propagating for [0^-]')
        engine0.propagate(path_tmp, ensembles[0], reverse=True)
    else:
        logger.debug('Not propagating for [0^-]')
        path_tmp.append(system)
    path0 = path_tmp.empty_path(maxlen=maxlen0)
    for phasepoint in reversed(path_tmp.phasepoints):
        path0.append(phasepoint)
    # Add second point from [0^+] at the end:
    logger.debug('Adding second point from [0^+]:')
    # Here we make a copy of the phase point, as we will update
    # the configuration and append it to the new path:
    phase_point = path_ensemble1.last_path.phasepoints[1].copy()
    logger.debug('Point is %s', phase_point)
    engine1.dump_phasepoint(phase_point, 'second')
    path0.append(phase_point)
    if path0.length == maxlen0:
        path0.status = 'BTX'
    elif path0.length < 3:
        path0.status = 'BTS'
    elif ('L' not in set(path_ensemble0.start_condition) and
          'L' in path0.check_interfaces(path_ensemble0.interfaces)[:2]):
        path0.status = '0-L'
    else:
        path0.status = 'ACC'

    # 2. Generate path for [0^+] from [0^-]:
    logger.debug('Creating path for [0^+] from [0^-]')
    # This path will be generated starting from the LAST point of [0^-] which
    # should be on the right side of the interface. We will also add the
    # SECOND LAST point from [0^-] which should be on the left side of the
    # interface, this is added after we have generated the path and we
    # save space for this point by letting maxlen = maxlen1-1 here:
    path_tmp = path0.empty_path(maxlen=maxlen1-1)
    # We start the generation from the LAST point:
    # Again, the copy below is not needed as the propagate
    # method will not alter the initial state.
    system = path_ensemble0.last_path.phasepoints[-1].copy()
    if allowed:
        logger.debug('Initial point is %s', system)
        ensembles[1]['system'] = system
        logger.debug('Propagating for [0^+]')
        engine1.propagate(path_tmp, ensembles[1], reverse=False)
        # Ok, now we need to just add the SECOND LAST point from [0^-] as
        # the first point for the path:
        path1 = path_tmp.empty_path(maxlen=maxlen1)
        phase_point = path_ensemble0.last_path.phasepoints[-2].copy()
        logger.debug('Add second last point: %s', phase_point)
        engine0.dump_phasepoint(phase_point, 'second_last')
        path1.append(phase_point)
        path1 += path_tmp  # Add rest of the path.
    else:
        path1 = path_tmp
        path1.append(system)
        logger.debug('Skipping propagating for [0^+] from L')

    if path_ensemble1.last_path.get_move() != 'ld':
        path0.set_move('s+')
    else:
        path0.set_move('ld')

    if path_ensemble0.last_path.get_move() != 'ld':
        path1.set_move('s-')
    else:
        path1.set_move('ld')
    if path1.length >= maxlen1:
        path1.status = 'FTX'
    elif path1.length < 3:
        path1.status = 'FTS'
    else:
        path1.status = 'ACC'
    logger.debug('Done with swap zero!')

    # Final checks:
    accept = path0.status == 'ACC' and path1.status == 'ACC'
    status = 'ACC' if accept else (path0.status if path0.status != 'ACC' else
                                   path1.status)
    # High Acceptance swap is required when Wire Fencing are used
    if accept and settings['tis'].get('high_accept', False):
        if 'wf' in ens_moves:
            accept, status = high_acc_swap([path1, path_ensemble1.last_path],
                                           ensembles[0]['rgen'],
                                           intf_w[0],
                                           intf_w[1],
                                           ens_moves)

    for i, path, path_ensemble, flag in ((0, path0, path_ensemble0, 's+'),
                                         (1, path1, path_ensemble1, 's-')):
        if not accept and path.status == 'ACC':
            path.status = status

        # These should be 1 unless length of paths equals 3.
        # This technicality is not yet fixed. (An issue is open as a reminder)

        ens_set = settings['ensemble'][i]
        move = ens_moves[i]
        path.weight = compute_weight(path, intf_w[i], move)\
            if (ens_set['tis'].get('high_accept', False) and
                move in ('wf', 'ss')) else 1

        if accept:
            path_ensemble.move_path_to_generate(path)
        else:
            logger.debug("Rejected swap path in [0^%s], %s", flag[:-1], status)
        path_ensemble.add_path_data(path, path.status, cycle=cycle)
        if cycle % ens_set.get('output', {}).get('restart-file', 1) == 0:
            write_ensemble_restart(ensembles[i], settings['ensemble'][i])

    return accept, (path0, path1), status


def repptis_swap(ensembles, idx, settings, cycle, inc_sh=True):
    """Perform a replica exchange move between adjacent PPTIS ensembles.

    First, propagation directions are randomly chosen for both paths, and it is
    checked whether the pathtypes (i.e. LMR, RMR, ...) are compatible. If not,
    the move is rejected. If the proposed directions are compatible with the
    pathtypes, the path segments in the overlapping region are cut, and they
    are propagated in the proposed directions.

    When compatible, the paths are created as follows:
    1) new [i^+-] path from old [(i+1)^+-] path. The path segment in the 'LM'
    region is cut out of the old [(i+1)^+-] path (with one point left of the
    L interface, and one point right of the M interface). This path segment is
    then extended by propagating the phasepoint left of the L interface in the
    proposed direction till a crossing condition of the [i^+-] interface occurs

    2) new [(i+1)^+-] path from old [i^+-] path. The path segment in the 'MR'
    region is cut out of the old [i^+-] path (with one point left of the M
    interface, and one point right of the R interface). This path segment is
    then extended by propagating the phasepoint right of the R interface in the
    proposed direction until a crossing condition of the [(i+1)^+-] interface
    occurs.

    Floating point precision problem (for GRO):
    The phase point to be shot from, is first dumped to a .gro file, where
    precision goes from ~9 decimals to 3 decimals. This impacts the order
    parameter of the dumped phasepoint, and if it was very closely located to
    an interface, it might even switch sides w.r.t. that interface. The problem
    has an easy fix. We include the shooting point in the overlap_path, such
    that the order parameter does not change. When we shoot from this point,
    the order parameter will still change due to precision decrease, and the
    first point of the propagated_path will be slightly different. We discard
    this point in the propagated_path, and keep the one from the overlap_path.
    Whichever shootpoint you keep, there will be a mismatch. The choice we make
    does not cause trouble with the swapping move. The above described behavior
    (i.e. the fix) is enabled with the paramter inc_sh set to True, being
    the standard behavior.

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the RETIS method.
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
        This is used for storing results for the simulation. It
        is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
        System is used here since we need access to the temperature
        and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
        The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
        The engine to use for propagating a path.

    idx : int
        The index of the ensemble that will be swapped.
    settings : dict
        This dict contains the settings for the RETIS method.
    cycle : integer
        The current cycle number.
    inc_sh : bool
        If True, the phasepoint that is used for shooting is included in the
        overlap path. If False, it will be included in the newly propagated
        path (with lower .gro precision).

    Returns
    -------
    out[0] : boolean
        Should the paths be accepted or not?
    out[1] : list of object like :py:class:`.PathBase`
        The trial paths.
    out[2] : string
        The status for the trial paths.

    """
    msg = f"Swapping: {ensembles[idx]['path_ensemble'].ensemble_name} <-> "
    msg += f"{ensembles[idx+1]['path_ensemble'].ensemble_name}"
    logger.info(msg)

    # if idx == 0, we perform a swap_zero move
    if idx == 0:
        # Since we allow RML paths in [0+-'], we need to do an extra check,
        # as this pathtype will result in a bogus swap...
        # NB: if we don't check for this, the move will be rejected anyways,
        # as the first propagated ph will be OOB for [0-], which results in
        # an "FTS" or "BTS" status. However, it's cleaner to reject the move
        # in advance...
        start, end, _, _ =\
            ensembles[1]['path_ensemble'].last_path.check_interfaces(
                ensembles[1]['path_ensemble'].interfaces)
        ptype = ""+start+"M"+end
        if ptype == "RML":
            msg = "Rejecting swap: RML in [0+-'] cannot be swapped with [0-]!"
            logger.info(msg)
            return reject_repptis_swap_wrong_prop(
                ensembles[0]['path_ensemble'], ensembles[1]['path_ensemble'],
                cycle, settings, idx, ensembles)
        return retis_swap_zero(ensembles, settings, cycle)

    # else we perform a repptis swap move
    path_ensemble0 = ensembles[idx]['path_ensemble']
    path_ensemble1 = ensembles[idx+1]['path_ensemble']

    # Choose propagation directions for the two paths, and check if the
    # proposed move is allowed or not
    allowed, propdir0, propdir1, path0_type, path1_type = \
        re_propagation_directions(ensembles, idx)

    # If not allowed, we immediately reject the swap
    if not allowed:
        return reject_repptis_swap_wrong_prop(path_ensemble0, path_ensemble1,
                                              cycle, settings, idx, ensembles)

    # If allowed, the swap move is attempted
    logger.info("Attempting swap.")
    logger.debug("Attempting swap of %s and %s paths with propdirs %s and %s.",
                 path0_type, path1_type, propdir0, propdir1)

    # First create path0_new, then path1_new.
    # We will immediately reject the move if path0_new is not valid.
    # -----------------------------------------------------------------------
    # Create new path for [i+-]
    maxlen0 = settings['ensemble'][idx]['tis']['maxlength']
    engine0 = ensembles[idx]['engine']
    # We distinguish between the part of the path that is copied (overlap,cut),
    # and the part that is propagated (prop).
    # First we cut:
    maxlen_cut0 = maxlen0 if inc_sh else maxlen0 - 1
    path_overlap0_new = path_ensemble0.last_path.empty_path(maxlen=maxlen_cut0)
    path0_new_shootpoint = \
        cut_overlap_phasepoints(path_ensemble1, propdir1, path_overlap0_new,
                                side='right', include_shootpoint=inc_sh)
    # Then create empty path to be filled with propagated points:
    # + 1 because the shootpoint will not be used in the propagated path.
    # (Thus, one point of prop_path will be discarded)
    path0_new = None
    path1_new = None
    maxlen_prop0_new = maxlen0 - path_overlap0_new.length + 1
    path_prop0_new = path_ensemble0.last_path.empty_path(
        maxlen=maxlen_prop0_new)
    # Set the system to the shootpoint:
    ensembles[idx]['system'] = path0_new_shootpoint
    logger.debug("Initial point is %s", ensembles[idx]['system'])
    # Propagate the path:
    engine0.propagate(path_prop0_new, ensembles[idx],
                      reverse=propdir1 == 'backwards')
    if inc_sh:  # We need to remove the shootpoint from the propagated path
        path_prop0_new.phasepoints.pop(0)
    # Append the propagated path to the overlap path:
    if propdir1 == 'backwards':
        path0_new = paste_paths(path_prop0_new, path_overlap0_new,
                                overlap=False, maxlen=maxlen0)
    elif propdir1 == 'forwards':
        path0_new = paste_paths(path_overlap0_new, path_prop0_new,
                                overlap=False, maxlen=maxlen0)
    # Check if the new path is valid:
    path0_new.status = set_swap_status(path0_new, propdir1, maxlen0)

    # If the new path is not valid, we stop and reject the entire swap move
    if path0_new.status != 'ACC':
        return reject_repptis_swap_half(path_ensemble0, path_ensemble1,
                                        path0_new, cycle, settings, idx,
                                        ensembles)
    # -----------------------------------------------------------------------
    # Create new path for [(i+1)+-]
    maxlen1 = settings['ensemble'][idx+1]['tis']['maxlength']
    engine1 = ensembles[idx+1]['engine']
    # First we cut:
    maxlen_cut1 = maxlen1 if inc_sh else maxlen1 - 1
    path_overlap1_new = path_ensemble1.last_path.empty_path(maxlen=maxlen_cut1)
    path1_new_shootpoint = \
        cut_overlap_phasepoints(path_ensemble0, propdir0, path_overlap1_new,
                                side='left', include_shootpoint=inc_sh)
    # Then create empty path to be filled with propagated points:
    # + 1 because the shootpoint will not be used in the propagated path.
    # (Thus, one point of prop_path will be discarded)
    maxlen_prop1_new = maxlen1 - path_overlap1_new.length + 1
    path_prop1_new = path_ensemble1.last_path.empty_path(
        maxlen=maxlen_prop1_new)
    # Set the system to the shootpoint:
    ensembles[idx+1]['system'] = path1_new_shootpoint
    logger.debug("Initial point is %s", ensembles[idx+1]['system'])
    # Propagate the path:
    engine1.propagate(path_prop1_new, ensembles[idx+1],
                      reverse=propdir0 == 'backwards')
    if inc_sh:  # We need to remove the shootpoint from the propagated path
        path_prop1_new.phasepoints.pop(0)
    # Append the propagated path to the overlap path:
    if propdir0 == 'backwards':
        path1_new = paste_paths(path_prop1_new, path_overlap1_new,
                                overlap=False, maxlen=maxlen1)
    elif propdir0 == 'forwards':
        path1_new = paste_paths(path_overlap1_new, path_prop1_new,
                                overlap=False, maxlen=maxlen1)
    # Check if the new path is valid:
    path1_new.status = set_swap_status(path1_new, propdir0, maxlen1)
    # -----------------------------------------------------------------------
    # Bookkeeping
    input_tuple = (path_ensemble0, path_ensemble1, path0_new, path1_new,
                   settings, idx, cycle, ensembles)
    return repptis_swap_bookkeeping(input_tuple)


def repptis_swap_bookkeeping(input_tuple):
    """Bookkeeping for the repptis swap move.

    If you reached here, then the first half of the swapping move was already
    successful. Here it is checked whether the second half of the move is
    successful or not. Afterward the final bookkeeping is performed.
    The input_tuple contains the parameters below, which are assigned at the
    beginning of the function.

    Parameters
    ----------
    input_tuple: tuple, it contains
        path_ensemble0 : object like :py:class:`.PathEnsemble`
            The path ensemble of the first trial path ([i+-]).
        path_ensemble1 : object like :py:class:`.PathEnsemble`
            The path ensemble of the second trial path ([(i+1)+-]).
        path0_new : object like :py:class:`.PathBase`
            The first trial path ([i+-]).
        path1_new : object like :py:class:`.PathBase`
            The second trial path ([(i+1)+-]).
        settings : dict
            This dict contains the settings for the REPPTIS method.
        idx : int
            The index of the ensemble that will be swapped with the next one.
        cycle : integer
            The current cycle number.
        ensembles : list of dictionaries of objects
            This is a list of the ensembles we are using in the REPPTIS method.

    Returns
    -------
    out[0] : boolean
        Should the paths be accepted or not?
    out[1] : list of object like :py:class:`.PathBase`
        The trial paths.
    out[2] : string
        The status for the trial paths.

    """
    # Get the input:
    path_ensemble0, path_ensemble1, path0_new, path1_new, settings, idx, \
        cycle, ensembles = input_tuple

    status = 'ACC'
    assert path0_new.status == 'ACC', "First half of swap should've been ACC."
    if path1_new.status != 'ACC':  # This can happen (FTX or BTX)
        status = path1_new.status

    for path_new, path_ensemble, flag in ((path0_new, path_ensemble0, 's+'),
                                          (path1_new, path_ensemble1, 's-')):
        path_new.status = status
        path_new.weight = 1  # NB: no high acceptance allowed for repptis
        path_new.set_move('ld' if path_ensemble.last_path.get_move() == 'ld'
                          else flag)

    accept = path0_new.status == 'ACC' and path1_new.status == 'ACC'
    # Here is where you would want to put 'plot_repptis_swap' for debugging
    # plot_repptis_swap(path_ensemble0, path_ensemble1, path0_new, path1_new,
    #                   idx, cycle, accept)
    if accept:
        path0_new = path_ensemble0.copy_path_to_generate(path0_new)
        path1_new = path_ensemble1.copy_path_to_generate(path1_new)

    for i, path_new, path_ensemble in ((0, path0_new, path_ensemble0),
                                       (1, path1_new, path_ensemble1)):
        path_ensemble.add_path_data(path_new, status, cycle=cycle)
        ens_set = settings['ensemble'][idx+i]
        if cycle % ens_set.get('output', {}).get('restart-file', 1) == 0:
            write_ensemble_restart(ensembles[idx+i], ens_set)
    return accept, (path0_new, path1_new), status


def set_swap_status(path, propdir, maxlen):
    """Set the status of a path after a swap-propagation in REPPTIS.

    Normally, the path extension should automatically result in an acceptable
    path of the adjacent ensemble, but there are two exceptions:
    1) The path is too long (exceeds maxlen): breaks detailed balance.
    2) The path is too short (length < 3): Not a path breaks detailed balance?

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        The path that was extended by a REPPTIS swap move.
    propdir : string
        The propagation direction of the path. Either 'forwards' or 'backwards'
    maxlen : int
        The maximum length of the path.

    Returns
    -------
    status : string
        The status of the path after the REPPTIS swap move.

    """
    # First we check if it's too long
    status_map = {'backwards': {True: 'BTX', False: 'ACC'},
                  'forwards': {True: 'FTX', False: 'ACC'}}
    status = status_map[propdir][path.length >= maxlen]
    # Then we check if it's too short. This is theoretically not possible,
    # as we already start from an overlap path that is by definition 2
    # phasepoints long (and we add another one with shooting). However, we
    # keep the check, just in case something terrible happened.
    assert path.length >= 3, "Pathlength < 3 is NOT possible in REPPTIS swap."

    return status


def reject_repptis_swap_wrong_prop(path_ensemble0, path_ensemble1, cycle,
                                   settings, idx, ensembles):
    """Reject REPPTIS swap for incompatible propagation directions.

    E.g. Trying to extend an LMR path of [i+-] into a [(i+1)+-] path by
    backwards propagation.

    Parameters
    ----------
    path_ensemble0 : object like :py:class:`.PathEnsemble`
        The path ensemble of the first trial path ([i+-]).
    path_ensemble1 : object like :py:class:`.PathEnsemble`
        The path ensemble of the second trial path ([(i+1)+-]).
    cycle : integer
        The current cycle number.
    settings : dict
        This dict contains the settings for the REPPTIS method.
    idx : int
        The index of the ensemble that will be swapped with the next one.
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the REPPTIS method.

    Returns
    -------
    out[0] : boolean
        Should the paths be accepted or not?
    out[1] : list of object like :py:class:`.PathBase`
        The trial paths.
    out[2] : string
        The status for the trial paths.

    """
    logger.info("Swap rejected: incompatible propagation directions")
    status = 'SWD'  # swap wrong direction
    accept = False
    # Make shallow copies:
    trial0 = copy.copy(path_ensemble1.last_path)
    trial1 = copy.copy(path_ensemble0.last_path)
    for trial, move in ((trial0, 's+'), (trial1, 's-')):
        trial.set_move('ld' if trial.get_move() == 'ld' else move)
        trial.weight = 1.0
    path_ensemble0.add_path_data(trial0, status, cycle=cycle)
    path_ensemble1.add_path_data(trial1, status, cycle=cycle)
    for i in (0, 1):
        ens_set = settings['ensemble'][idx+i]
        if cycle % ens_set.get('output', {}).get('restart-file', 1) == 0:
            write_ensemble_restart(ensembles[idx+i], ens_set)
    return accept, (trial0, trial1), status


def reject_repptis_swap_half(path_ensemble0, path_ensemble1, path0_new, cycle,
                             settings, idx, ensembles):
    """Bookkeeping when the first REPPTIS extension is not acceptable.

    If the first extension of the REPPTIS swap move does not result in an
    acceptable path for the [(i+1)+-] ensemble, we do not waste computational
    time on the second extension. We immediately reject the move.

    Parameters
    ----------
    path_ensemble0 : object like :py:class:`.PathEnsemble`
        The path ensemble of the first trial path ([i+-]).
    path_ensemble1 : object like :py:class:`.PathEnsemble`
        The path ensemble of the second trial path ([(i+1)+-]).
    path0_new : object like :py:class:`.PathBase`
        The first trial path ([i+-]).
    cycle : integer
        The current cycle number.
    settings : dict
        This dict contains the settings for the REPPTIS method.
    idx : int
        The index of the ensemble that will be swapped with the next one.
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the REPPTIS method.

    Returns
    -------
    out[0] : boolean
        Should the paths be accepted or not?
    out[1] : list of object like :py:class:`.PathBase`
        The trial paths.
    out[2] : string
        The status for the trial paths.

    """
    logger.info("Swap rejected: first path not accepted")
    status = 'SWH'  # swap rejected
    accept = False
    # Make shallow copies:
    trial0 = copy.copy(path0_new)
    trial1 = copy.copy(path_ensemble0.last_path)
    for trial, move in ((trial0, 's+'), (trial1, 's-')):
        trial.set_move('ld' if trial.get_move() == 'ld' else move)
        trial.weight = 1.0
    path_ensemble0.add_path_data(trial0, status, cycle=cycle)
    path_ensemble1.add_path_data(trial1, status, cycle=cycle)
    for i in (0, 1):
        ens_set = settings['ensemble'][idx+i]
        if cycle % ens_set.get('output', {}).get('restart-file', 1) == 0:
            write_ensemble_restart(ensembles[idx+i], ens_set)
    return accept, (trial0, trial1), status


def cut_overlap_phasepoints(path_ensemble, propdir, overlap_path, side,
                            include_shootpoint=True):
    """Cut path in overlapping region of adjacent ensembles.

    Cuts the phasepoints in the overlapping region of two adjacent PPTIS
    ensembles [i^+] and [(i+1)^-]. The overlapping region is [L,M] if
    we are cutting from an [(i+1)^+-] path (denoted by side = 'right'), and
    [M,R] if we are cutting from an [i^+-] path (denoted by side = 'left).
    The propagation direction is the direction in which the cut path is to be
    propagated.

    Parameters
    ----------
    path_ensemble: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
    propdir : string
        This is either 'forwards' or 'backwards', and denotes the direction
        in which the cut path is to be propagated.
    overlap_path : object like :py:class:`.Path`
        The path that is to be cut.
    side : string
        This is either 'left' or 'right', and denotes the side from which the
        path originates. left: [i^+-] path to be propagated into [(i+1)^+-]
        right: [(i+1)^+-] path to be propagated into [i^+-]
    include_shootpoint : bool
        If True, the shootpoint of the path is included in the cut path.

    Returns
    -------
    shoot_point : object like :py:class:`.PhasePoint`
        The phasepoint from which the cut path is to be propagated.

    """
    # intf_M defines when to stop cutting phasepoints
    # For now, this is equivalent for all PPTIS ensembles, but might change
    # in the future.
    assert side in ['right', 'left']

    intf_m = path_ensemble.interfaces[0] if path_ensemble.ensemble_number == 1\
        else path_ensemble.interfaces[1]

    stop_idx = 1  # to not include shootpoint in the cut path loop

    # Different behavior depending on the side from which the path originates,
    # and the propagation direction. Note that, for the paths. Note that, if
    # the propagation direction is forwards, we reverse the order in which we
    # cut the phasepoints. We do not re-reverse this in the end, because the
    # function 'paste_paths' will be used. As this cut-out path is the backward
    # part of the merged paths, it will be reversed in that function.
    # For the backwards propagation, the cut-out path is the forwards part of
    # the merged paths, and will not be reversed in 'paste_paths' (or here).
    if side == 'right':
        if propdir == 'backwards':
            shoot_point = path_ensemble.last_path.phasepoints[0].copy()
            if include_shootpoint:
                overlap_path.append(shoot_point.copy())
            phase_points = path_ensemble.last_path.phasepoints[stop_idx:]
            for phase_point in phase_points:
                overlap_path.append(phase_point.copy())
                if phase_point.order[0] >= intf_m:  # if>=M
                    break
        elif propdir == 'forwards':
            shoot_point = path_ensemble.last_path.phasepoints[-1].copy()
            if include_shootpoint:
                overlap_path.append(shoot_point.copy())
            phase_points = path_ensemble.last_path.phasepoints[:-stop_idx]
            for phase_point in reversed(phase_points):
                overlap_path.append(phase_point.copy())
                if phase_point.order[0] >= intf_m:  # if>=M
                    break

    elif side == 'left':
        if propdir == 'backwards':
            shoot_point = path_ensemble.last_path.phasepoints[0].copy()
            if include_shootpoint:
                overlap_path.append(shoot_point.copy())
            phase_points = path_ensemble.last_path.phasepoints[stop_idx:]
            for phase_point in phase_points:
                overlap_path.append(phase_point.copy())
                if phase_point.order[0] <= intf_m:  # if<=M
                    break
        elif propdir == 'forwards':
            shoot_point = path_ensemble.last_path.phasepoints[-1].copy()
            if include_shootpoint:
                overlap_path.append(shoot_point.copy())
            phase_points = path_ensemble.last_path.phasepoints[:-stop_idx]
            for phase_point in reversed(phase_points):
                overlap_path.append(phase_point.copy())
                if phase_point.order[0] <= intf_m:  # if<=M
                    break

    return shoot_point


def fix_directions(ensembles, idx):  # pragma: no cover
    """Propose acceptable propagation direcs for [i^+-] and [(i+1)+-] paths.

    If no acceptable directions are found, a rejection is returned.
    Note that this will **break detailed balance**.

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the RETIS method.
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    idx : integer
        Index of the ensemble, which will be swapped with the (idx+1) ensemble.

    Returns
    -------
    accepted : bool
        Whether the proposed directions can result in an acceptable swap.
    propdir0 : string
        The proposed propagation direction for the [i^+] path. This is either
        'forwards' or 'backwards', or None if no acceptable direction was found
    propdir1 : string
        The proposed propagation direction for the [(i+1)^-] path. This is
        either 'forwards' or 'backwards', or None if no acceptable direction
        was found.
    path0_type : string
        The type of the [i^+] path. This is either 'LMR', 'RMR', 'LML' or 'RML'
    path1_type : string
        The type of the [(i+1)^-] path. This is either 'LMR', 'RMR', 'LML' or
        'RML'.

    Notes
    -----
    This function does **not** obey detailed balance (DB). Used for debugging.
    It could allow detailed balance, but then the path ensemble definitions
    should be changed (weight dependent on path type!)

    """
    path_ensemble0 = ensembles[idx]['path_ensemble']
    path_ensemble1 = ensembles[idx+1]['path_ensemble']

    # Get the start and end points of the paths (L or R)
    path0_old_start, path0_old_end, _, _ = \
        path_ensemble0.last_path.check_interfaces(path_ensemble0.interfaces)
    path1_old_start, path1_old_end, _, _ = \
        path_ensemble1.last_path.check_interfaces(path_ensemble1.interfaces)

    # Define the path types, for nice output
    path0_type = ""+path0_old_start+"M"+path0_old_end
    path1_type = ""+path1_old_start+"M"+path1_old_end

    # Define the allowed propagation directions
    propdirs = {'leftright': {'LMR': 'forwards',
                              'RMR': 'forwards',  # backwards could work too
                              'LML': None,
                              'RML': 'backwards'},
                'rightleft': {'LMR': 'backwards',
                              'RMR': None,
                              'LML': 'backwards',  # forwards could work too
                              'RML': 'forwards'}}

    # allocate directions
    propdir0 = propdirs['leftright'][path0_type]
    propdir1 = propdirs['rightleft'][path1_type]

    # Check whether it is acceptable
    accepted = propdir0 in ('forwards', 'backwards') and \
        propdir1 in ('forwards', 'backwards')

    # Log the move
    msg = "Proposed swap move:\n"
    msg += f"{path_ensemble0.ensemble_name} <-->" + \
           f"{path_ensemble1.ensemble_name}\n" + \
           f"{path0_type} <-----> {path1_type}\n" + \
           f"{propdir0}<>{propdir1}"
    logger.info(msg)

    return accepted, propdir0, propdir1, path0_type, path1_type


def re_propagation_directions(ensembles, idx):
    """Proposes propagation directions for the [i^+-] and [(i+1)^+-] paths.

    And checks whether this results in an acceptable swapping move.

    Parameters
    ----------
    ensembles : list of dictionaries of objects
        This is a list of the ensembles we are using in the RETIS method.
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is used for storing results for the simulation. It
          is also used for defining the interfaces for this simulation.
        * `system`: object like :py:class:`.System`
          System is used here since we need access to the temperature
          and to the particle list.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for calculating the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.

    idx : integer
        Index of the ensemble, which will be swapped with the (idx+1) ensemble.

    Returns
    -------
    allowed : bool
        Whether the proposed directions can result in an acceptable swap.
    propdir0 : string
        The proposed propagation direction for the [i^+-] path. This is either
        'forwards' or 'backwards', or None if no acceptable direction was found
    propdir1 : string
        The proposed propagation direction for the [(i+1)^+-] path. This is
        either 'forwards' or 'backwards', or None if no acceptable direction
        was found.
    path0_type : string
        The type of the [i^+-] path. This is 'LMR', 'RMR', 'LML' or 'RML'.
    path1_type : string
        The type of the [(i+1)^+-] path. This is either 'LMR', 'RMR', 'LML' or
        'RML'.

    """
    # Set path ensembles
    path_ensemble0 = ensembles[idx]['path_ensemble']
    path_ensemble1 = ensembles[idx+1]['path_ensemble']

    # Check the start and end points of the paths (L or R)
    path0_old_start, path0_old_end, _, _ = \
        path_ensemble0.last_path.check_interfaces(path_ensemble0.interfaces)
    path1_old_start, path1_old_end, _, _ = \
        path_ensemble1.last_path.check_interfaces(path_ensemble1.interfaces)

    # Define the path types, for nice output
    path0_type = ""+path0_old_start+"M"+path0_old_end
    path1_type = ""+path1_old_start+"M"+path1_old_end

    # Pick two random numbers rand0 and rand1 between [0,1].
    # If rand0 < 0.5: attempt backward propagation for path0. Otherwise forward
    # If rand1 < 0.5: attempt backward propagation for path1. Otherwise forward

    # For now, we will just pick a random number for each path from the
    # random number generator of the 0 ensemble, because there are seed issues.
    idx1 = int(round((ensembles[0]['rgen'].rand())[0]) + 0.1)
    idx2 = int(round((ensembles[0]['rgen'].rand())[0]) + 0.1)
    propdir0 = ['backwards', 'forwards'][idx1]
    propdir1 = ['backwards', 'forwards'][idx2]

    # Check if the proposed move is allowed
    allow = {'leftright': {'L': False, 'R': True},
             'rightleft': {'L': True, 'R': False}}

    allowed0 = allow['leftright'][path0_old_start if propdir0 == 'backwards'
                                  else path0_old_end]
    allowed1 = allow['rightleft'][path1_old_start if propdir1 == 'backwards'
                                  else path1_old_end]

    # Both paths must be allowed to begin the swapping move
    allowed = allowed0 and allowed1

    # Log the move
    msg = "Proposed swap move:\n"
    msg += f"{path_ensemble0.ensemble_name} <-->" + \
           f"{path_ensemble1.ensemble_name}\n" + \
           f"{path0_type} <-----> {path1_type}\n" + \
           f"{propdir0}<>{propdir1}\n" + \
           f"Allowed: {allowed}"
    logger.info(msg)

    return allowed, propdir0, propdir1, path0_type, path1_type


def high_acc_swap(paths, rgen, intf0, intf1, ens_moves):
    """Accept or Reject a swap move using the High Acceptance weights.

    Parameters
    ----------
    paths: list of object like :py:class:`.PathBase`
        The path in the LOWER and UPPER ensemble to exchange.
    rgen : object like :py:class:`.RandomGenerator`
        This is a random generator.
    intf0: list of float
        The interfaces of the LOWER ensemble.
    intf1: list of float
        The interfaces of the HIGHER ensemble.
    ens_moves: list of string
        The moves used in the two ensembles.

    Returns
    -------
    out[0] : boolean
        True if the move should be accepted.

    Notes
    -----
     -  This function is needed only when paths generated via Wire Fencing or
        Stone Skipping are involved.
      - In the case that a path bears a flag 'ld', the swap is accepted,
            but the flag will be unchanged.

    """
    # Crossing before the move
    c1_old = compute_weight(paths[0], intf0, ens_moves[0])
    c2_old = compute_weight(paths[1], intf1, ens_moves[1])
    # Crossing if the move would be accepted
    c1_new = compute_weight(paths[1], intf0, ens_moves[0])
    c2_new = compute_weight(paths[0], intf1, ens_moves[1])
    if c1_old == 0 or c2_old == 0:
        logger.warning("div_by_zero. c1_old, c2_old, ens_moves: [%i,%i], %s",
                       c1_old, c2_old, str(ens_moves))
        p_swap_acc = 1
    else:
        p_swap_acc = c1_new*c2_new/(c1_old*c2_old)

    # Finally, randomly decide to accept or not:
    if rgen.rand() < p_swap_acc:
        return True, 'ACC'  # Accepted

    return False, 'HAS'  # Rejected



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the oder_parameter methods in pyretis.pyvisa.orderparam_density."""
import logging
import os
import warnings
import unittest
from io import StringIO
import numpy as np
import pandas as pd
from subprocess import DEVNULL, STDOUT, check_call
from pyretis.pyvisa.orderparam_density import (PathDensity,
                                               PathVisualize,
                                               remove_nan,
                                               pyvisa_compress)
from unittest import mock

logging.disable(logging.CRITICAL)
warnings.filterwarnings('ignore', category=pd.io.pytables.PerformanceWarning)

BASE = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                    'test_simulation_dir')
BASE_ZIP = os.path.join(BASE, 'pyvisa_compressed_data.hdf5.zip')
INPUTFILE = os.path.join(BASE, 'input.rst')

CORRECT = {
    'path': ['000', '001', '002', '003', '004', '005'],
    'op_labels': ['op1', 'op2'],
    'energy_labels': ['time', 'cycle', 'potE', 'kinE', 'totE'],
    'interfaces': [-0.9, -0.8, -0.7, -0.5, -0.4, -0.3, 1.0],
    'num_op': 2
}

CORRECT_LENGTHS_SLOW = {
    '000 A': [968, 968, 968, 968, 3872, 3872, 3872, 3872],
    '000 R': [0, 0, 0, 0, 1500, 1500, 1500, 1500],
    '001 A': [247, 247, 247, 247, 1235, 1235, 1235, 1235],
    '001 R': [454, 454, 454, 454, 2719, 2719, 2719, 2719],
    '002 A': [247, 247, 247, 247, 2223, 2223, 2223, 2223],
    '002 R': [454, 454, 454, 454, 2719, 2719, 2719, 2719],
    '003 A': [247, 247, 247, 247, 2223, 2223, 2223, 2223],
    '003 R': [454, 454, 454, 454, 2719, 2719, 2719, 2719],
}

LENGTHS = {'000': {0: [484], 1: [500], 2: [500], 3: [500],
                   4: [484], 5: [484],
                   6: [484], 7: [484], 8: [484], 9: [484], 10: [484]},
           '001': {0: [247], 1: [500], 2: [492], 3: [247],
                   4: [454], 5: [247], 6: [247],
                   7: [247], 8: [454], 9: [365], 10: [454]},
           '002': {0: [247], 1: [500], 2: [492, 500, 492],
                   3: [247, 247], 4: [454, 454],
                   5: [247], 6: [247],
                   7: [247], 8: [454], 9: [365], 10: [454]},
           '005': {0: [17], 1: [19], 2: [19]}}

STATUS = {'005': {0: 'ACC', 1: 'ACC', 2: 'NCR',
                  3: 'NCR', 4: 'ACC', 5: 'ACC', 6: 'ACC'}}
MOVES = {'005': {0: 'ki', 1: 'tr', 2: 's+',
                 3: 's+', 4: 'tr', 5: 's-', 6: 'tr'}}


def init_pathdensity(base=BASE, ifile=INPUTFILE):
    """Initialize PathDensity with input file."""
    with mock.patch('sys.stdout', new=StringIO()):
        data_dict = PathDensity(base, ifile)
    return data_dict


def init_pathvisualize(base=BASE, pfile=INPUTFILE):
    """Initialize PathVisualize."""
    with mock.patch('sys.stdout', new=StringIO()):
        visual = PathVisualize(base, pfile=pfile)
    return visual


def compare_keys(dict1, correct):
    """Test if given dictionary contains the correct set of keys."""
    check = True
    for key in correct:
        check = True if key in dict1 else False
    return check


class TestMethods(unittest.TestCase):
    """Testing class of pyretis.pyvisa.orderparam_density."""

    def test_PathVisualize_blank(self):
        """Testing initiate PathVisualize."""
        mori = PathVisualize()
        mori.pfile = os.path.join(BASE, 'traj.txt')
        with self.assertRaises(ValueError) as err:
            mori.load_whatever()
            self.assertIn('not recognised', err.exception)

    def test_PathDensity_blank(self):
        """Testing initiate PathDensity with a compressed input file."""
        with self.assertRaises(FileNotFoundError) as err:
            init_pathdensity(ifile='blank.hdf5')
            self.assertIn('File blank.hdf5 does not exist.', err.exception)

    def test_PathDensity(self):
        """Testing the PathDensity class creation from file."""
        ipd = init_pathdensity(BASE, INPUTFILE)
        self.assertEqual(ipd.iofile, INPUTFILE)
        self.assertTrue(compare_keys(ipd.infos, CORRECT.keys()))
        self.assertTrue(compare_keys(ipd.infos['ensemble_names'],
                                     CORRECT['path']))
        self.assertTrue(compare_keys(ipd.infos['op_labels'],
                                     CORRECT['op_labels']))
        self.assertTrue(compare_keys(ipd.infos['energy_labels'],
                                     CORRECT['energy_labels']))
        self.assertTrue(compare_keys(ipd.infos['interfaces'],
                                     CORRECT['interfaces']))
        self.assertEqual(ipd.infos['num_op'], CORRECT['num_op'])

        re_d = PathDensity(basepath='.', iofile=ipd)
        self.assertIsInstance(re_d, PathDensity)
        # test minor functionalities
        re_d.get_traj_energy('no_energy')

    def test_PathDensity_instance(self):
        """Test comparing PathDensity object."""
        pe_d = init_pathdensity()
        pe_e = init_pathdensity()
        self.assertEqual(pe_e.infos, pe_d.infos)
        self.assertEqual(pe_e.traj_dict, pe_d.traj_dict)

    def test_walk_Dirs(self):
        """Testing the walk_Dirs function of PathDensity."""
        pe_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            pe_d.walk_dirs()

    def test_pyvisa_compress(self):
        """Testing the compression of input files."""
        pyvisa_dict_hdf5 = {'only_order': True}
        with mock.patch('sys.stdout', new=StringIO()):
            pyvisa_compress(BASE, INPUTFILE,
                            pyvisa_dict_hdf5)
        self.assertTrue(os.path.isfile('pyvisa_compressed_data.hdf5.zip'))

        with self.assertRaises(ImportError) as err:
            pyvisa_compress('.', 'pyvisa_compressed_data.hdf5.zip', {})
            self.assertTrue('Cannot compress an already' in err.exception)
        check_call(['unzip', 'pyvisa_compressed_data.hdf5.zip'],
                   stdout=DEVNULL, stderr=STDOUT)

        with mock.patch('sys.stdout', new=StringIO()):
            re_d = PathDensity('.', 'pyvisa_compressed_data.hdf5')
        self.assertIsInstance(re_d, PathDensity)
        os.remove('pyvisa_compressed_data.hdf5.zip')
        os.remove('pyvisa_compressed_data.hdf5')

    def test_hdf5(self):
        """Testing the saving of a hdf5 file."""
        pe_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            pe_d.walk_dirs()
            # This is to remove the Deprecation Warning due to pandas
            with mock.patch('sys.stderr', new=StringIO()):
                pe_d.hdf5_data()
        self.assertTrue(os.path.isfile('pyvisa_compressed_data.hdf5.zip'))
        os.remove('pyvisa_compressed_data.hdf5.zip')

    def test_hdf5ing_and_loading(self):
        """Test for hdf5ing data to file and load with PathVisualize."""
        re_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            re_d.walk_dirs()
            # This is to remove the Deprecation Warning due to pandas.
            with mock.patch('sys.stderr', new=StringIO()):
                re_d.hdf5_data()
        self.assertTrue(os.path.isfile('pyvisa_compressed_data.hdf5.zip'))
        init_pathvisualize('pyvisa_compressed_data.hdf5.zip')
        os.rename('pyvisa_compressed_data.hdf5.zip', 'myhdf.hdf6.zip')
        with mock.patch('sys.stdout', new=StringIO()):
            with self.assertRaises(TypeError) as err:
                init_pathvisualize('myhdf.hdf6.zip')
                self.assertIn('recognised', err)
        os.remove('myhdf.hdf6.zip')

    def test_get_trajectory_op_length(self):
        """Test for correct lengths of collected trajectories."""
        re_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            re_d.walk_dirs()
        for ens in LENGTHS.keys():
            for cyc in LENGTHS[ens]:
                for idx in range(0, len(LENGTHS[ens][cyc])):
                    self.assertEqual(LENGTHS[ens][cyc][0],
                                     re_d.traj_dict[ens][cyc].info['length'])

    def test_load_trajectory(self):
        """Test for loading data."""
        return
        re_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            re_d.walk_dirs()
            with mock.patch('sys.stderr', new=StringIO()):
                re_d.hdf5_data()
        re_w = init_pathvisualize('pyvisa_compressed_data.hdf5.zip')
        os.remove(BASE_ZIP)

        criteria = {'x': 'op1', 'y': 'op2', 'z': 'time',
                    'cycles': re_d.infos['cycles'],
                    'weight': False}

        for ens in re_d.infos['ensemble_names']:
            x, y, z = [], [], []

            for cyc in range(criteria['cycles'][0], criteria['cycles'][1] + 1):
                if cyc not in re_d.traj_dict[ens].keys():
                    continue
                else:
                    traj = re_d.traj_dict[ens][cyc]
                    for xyz in {'x': x, 'y': y, 'z': z}.items():
                        if criteria[xyz[0]] == 'time':
                            xyz[1].extend(range(0, traj.info['length']))
                        elif criteria[xyz[0]] == 'cycle':
                            xyz[1].extend([traj.info['cycle']] *
                                          traj.info['length'])
                        elif criteria[xyz[0]] == 'None':
                            xyz[1].extend([1] * traj.info['length'])
                        elif criteria[xyz[0]] in traj.frames.columns:
                            xyz[1].extend(traj.frames[criteria[xyz[0]]])

            criteria['ensemble_name'] = ens
            x2, y2, z2, _ = re_w.load_traj(criteria)

            self.assertEqual(x, x2)
            self.assertEqual(y, y2)
            self.assertEqual(z, z2)

    def test_move_status(self):
        """Test for move status."""
        pe_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            pe_d.walk_dirs()
        for cyc in pe_d.traj_dict['005'].keys():
            traj = pe_d.traj_dict['005'][cyc]
            self.assertEqual(traj.info['status'], STATUS['005'][cyc])
            self.assertEqual(traj.info['MC-move'], MOVES['005'][cyc])

    def test_load_traj_acc_rej_both_hdf5(self):
        """Load ACC, REJ and BOTH trajectories from hdf5."""
        pe_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            pe_d.walk_dirs()
            pe_d.hdf5_data()
        pe_w = init_pathvisualize('.', 'pyvisa_compressed_data.hdf5.zip')
        os.remove('pyvisa_compressed_data.hdf5.zip')
        c_acc = {'status': 'ACC', 'x': 'op1', 'y': 'potE',
                 'z': 'time', 'cycles': pe_d.infos['cycles'],
                 'ensemble_name': '005', 'weight': False}
        c_rej = {'status': 'REJ', 'x': 'op2', 'y': 'totE',
                 'z': 'cycle', 'cycles': pe_d.infos['cycles'],
                 'ensemble_name': '005', 'weight': False}
        c_both = {'status': 'BOTH', 'x': 'op1', 'y': 'kinE',
                  'z': 'None', 'cycles': pe_d.infos['cycles'],
                  'ensemble_name': '005', 'weight': False}
        x_a, y_a, z_a = [], [], []
        x_r, y_r, z_r = [], [], []
        x_b, y_b, z_b = [], [], []
        for cycle in pe_d.traj_dict['005'].keys():
            traj = pe_d.traj_dict['005'][cycle]
            if traj.info['status'] == 'ACC':
                x_a.extend(traj.frames['op1'])
                y_a.extend(traj.frames['potE'])
                z_a.extend(range(0, traj.info['length']))
            else:
                x_r.extend(traj.frames['op2'])
                y_r.extend(traj.frames['totE'])
                z_r.extend([traj.info['cycle']] *
                           traj.info['length'])

            x_b.extend(traj.frames['op1'])
            y_b.extend(traj.frames['kinE'])
            z_b.extend([1] * traj.info['length'])

        x2_a, y2_a, z2_a, _ = pe_w.load_traj(c_acc)
        # Use assert_array_equal which correctly handles NaN == NaN
        np.testing.assert_array_equal(x_a, x2_a)
        np.testing.assert_array_equal(y_a, y2_a)
        np.testing.assert_array_equal(z_a, z2_a)

        x2_r, y2_r, z2_r, _ = pe_w.load_traj(c_rej)
        np.testing.assert_array_equal(x_r, x2_r)
        np.testing.assert_array_equal(y_r, y2_r)
        np.testing.assert_array_equal(z_r, z2_r)

        x2_b, y2_b, z2_b, _ = pe_w.load_traj(c_both)
        np.testing.assert_array_equal(x_b, x2_b)
        np.testing.assert_array_equal(y_b, y2_b)
        np.testing.assert_array_equal(z_b, z2_b)

    def test_no_op(self):
        """Remove Trajectory with no OP values."""
        pe_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            pe_d.walk_dirs()
        self.assertTrue(3 not in pe_d.traj_dict['005'].keys())
        self.assertTrue(0 in pe_d.traj_dict['005'].keys())

    def test_more_ops(self):
        """Remove Trajectory with no OP values."""
        pe_d = init_pathdensity()
        with mock.patch('sys.stdout', new=StringIO()):
            pe_d.walk_dirs()
        self.assertEqual(5, len(pe_d.traj_dict['005'][4].frames.columns))
        self.assertEqual(5, len(pe_d.traj_dict['005'][5].frames.columns))
        self.assertEqual(2, len(pe_d.traj_dict['004'][1].frames.columns))

    def test_remove_nan(self):
        """Test for stability in the most dramatic case.

        Note that the remove_nan function alters the input in place.

        """
        frames = {'0': {'op1': float('nan'), 'op2': 0},
                  '1': {'op1': float('nan'), 'op2': float('nan')},
                  '2': {'op1': float('nan'), 'op2': float('nan')},
                  '3': {'op1': float('nan'), 'op2': float('nan')}}
        data = pd.DataFrame.from_dict(frames, orient='index')
        data['op3'] = [float('nan'), float('nan'), float('nan'), 4]
        data['op4'] = [0, 2, float('nan'), 4]

        remove_nan(data)
        # Nothing can be fixed here.
        self.assertFalse(data['op1'].iloc[0] * 0 == 0)
        # Nothing can be fixed here and the start shall remain the same.
        self.assertEqual(data['op2'].iloc[0], 0)
        self.assertFalse(data['op2'].iloc[3] * 0 == 0)
        # Test nan at the beginning, it should be fixed.
        self.assertEqual(data['op3'].iloc[0], 4)
        self.assertEqual(data['op3'].iloc[3], 4)
        # Test nan disappearing (most common case).
        self.assertEqual(data['op4'].iloc[2], 4)

        # Test that the function works with normal lists also.
        bad = [0, float('nan'), 1]
        not_bad = [0, 1, 1]
        remove_nan([bad])
        self.assertEqual(bad, not_bad)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Common plotting methods used by the LAMMPS engine tests."""
from math import ceil
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec
from pyretis.inout import print_to_screen


def plot_compare(data_sets, ylabels):
    """Plot data for comparison.

    Parameters
    ----------
    data_sets : list of list of tuples
        The list are of the form `[(x, y, label), (x, y, label)]`.
    ylabels : list of strings
        The y-labels to use for each data set.

    """
    markers = ['o', 'x', '1']
    fig = plt.figure()
    if len(data_sets) < 3:
        ncol = 1
    else:
        ncol = 2
    grid = make_grid(ncol, len(data_sets))
    for i, data in enumerate(data_sets):
        row, col = divmod(i, ncol)
        axi = fig.add_subplot(grid[row, col])
        for j, dataj in enumerate(data):
            line, = axi.plot(dataj[0], dataj[1], ls='--', lw=1, alpha=0.5)
            axi.scatter(dataj[0], dataj[1], marker=markers[j],
                        s=100, alpha=0.7, label=dataj[2],
                        color=line.get_color())
        axi.set_xlabel('Step')
        axi.set_ylabel(ylabels[i])
        axi.legend()
    fig.tight_layout()


def make_grid(ncol, nsets):
    """Create a grid for plotting some data sets.

    Parameters
    ----------
    ncol : int
        The number of columns we want to plot in.
    nsets : int
        The number of data sets to plot.

    """
    nrow = ceil(nsets / ncol)
    grid = GridSpec(nrow, ncol)
    return grid


def plot_xy(data_sets):
    """Plot x-data vs y-data.

    This is intended to plot two data sets we expect to be equal,
    in order to visualize differences between the two data sets.

    Parameters
    ----------
    data_sets : list of tuples
        The tuples are on form `(x, y, xlabel, ylabel)`.

    """
    fig = plt.figure()
    ncol = 2
    grid = make_grid(ncol, len(data_sets))
    for i, data in enumerate(data_sets):
        row, col = divmod(i, ncol)
        axi = fig.add_subplot(grid[row, col])
        axi.scatter(data[0], data[1], marker='o', s=90, alpha=0.5)
        axi.set_xlabel(data[2])
        axi.set_ylabel(data[3])
        minxy = min(min(data[0]), min(data[1]))
        maxxy = max(max(data[0]), max(data[1]))
        axi.plot([minxy, maxxy], [minxy, maxxy], ls=':',
                 color='#262626', alpha=0.5, lw=2)
        axi.set_aspect('equal')
        axi.set_xlim([minxy, maxxy])
        axi.set_ylim([minxy, maxxy])
        if not np.allclose(data[0], data[1]):
            print_to_screen(
                f'Comparison: "{data[2]}" & "{data[3]}" failed.',
                level='error',
            )
        else:
            print_to_screen(
                f'Comparison: "{data[2]}" & "{data[3]}" is OK.',
                level='success',
            )
    fig.tight_layout()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file contains a class for a generic force field.

This module defines the class used for representing a force field.
The forcefield class is built up of potential functions.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ForceField (:py:class:`.ForceField`)
    A class representing a generic Force Field.
"""
import logging
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['ForceField']


class ForceField:
    """Represents a generic force field.

    This class described a generic Force Field.
    A force field is assumed to consist of a number of potential
    functions with parameters.

    Attributes
    ----------
    desc : string
        Description of the force field.
    potential : list
        The potential functions that the force field is built up from.
    params : list
        The parameters for the corresponding potential functions.

    """

    def __init__(self, desc, potential=None, params=None):
        """Initialise the force field object.

        Parameters
        ----------
        desc : string
            Description of the force field.
        potential : list, optional
            Potential functions that the force field is built up from.
        params : list, optional
            Parameters for the potential(s). If too few parameters are
            given, we will just assume a `None`.

        """
        self.desc = desc
        self.potential = []
        self.params = []
        if potential is not None:
            if params is None:
                for pot in potential:
                    self.add_potential(pot)
            else:
                for i, pot in enumerate(potential):
                    try:
                        param = params[i]
                    except IndexError:
                        param = None
                        msg = 'No parameters given for potential no. {} ({})'
                        msgtxt = msg.format(i, pot)
                        logger.warning(msgtxt)
                    self.add_potential(pot, parameters=param)

    def add_potential(self, potential, parameters=None):
        """Add a potential with parameters to the force field.

        Parameters
        ----------
        potential : object like :py:class:`.PotentialFunction`
            Potential function to add.
        parameters : dict, optional
            Parameters for the potential.

        Returns
        -------
        out : boolean
            Returns `True` and updates `self.potential` and
            `self.params` if the potential was added. Returns
            `False` otherwise.

        """
        if potential is None:
            msg = ('Trying to add empty potential to force field.\n'
                   'This was ignored -- please check your settings.')
            logger.warning(msg)
            return False
        self.potential.append(potential)
        if parameters is not None:
            potential.set_parameters(parameters)
        self.params.append(parameters)
        return True

    def remove_potential(self, potential):
        """Remove a selected potential from the force field.

        Parameters
        ----------
        potential : object like :py:class:`.PotentialFunction`
            The potential function to remove.

        Returns
        -------
        out : None or tuple
            Returns `None` if not potential was removed, otherwise it
            will return the removed potential and its parameters.

        """
        if potential in self.potential:
            idx = self.potential.index(potential)
            potrm = self.potential.pop(idx)
            paramrm = self.params.pop(idx)
            return potrm, paramrm
        logger.warning('Potential not found in the force field functions')
        return None, None

    def update_potential_parameters(self, potential, params):
        """Update the potential parameters of the given potential function.

        Parameters
        ----------
        potential : object like :py:class:`.PotentialFunction`
            Potential to update. Should be in `self.potential`.
        params : dict
            The new parameters to set.

        Returns
        -------
        out : None
            Returns `None` but will update parameters of the selected
            potential and modify the corresponding `self.params`.

        """
        if potential in self.potential:
            potential.set_parameters(params)
            self.params[self.potential.index(potential)] = params
        else:
            logger.warning('Unknow potential. Will not update!')

    def evaluate_force(self, system):
        """Evaluate the force on the particles.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the forces in.

        Returns
        -------
        out[0] : numpy.array
            The forces on the particles.
        out[1] : numpy.array
            The virial.

        """
        force = None
        virial = None
        for pot in self.potential:
            if force is None or virial is None:
                force, virial = pot.force(system)
            else:
                forcei, viriali = pot.force(system)
                force += forcei
                virial += viriali
        return force, virial

    def evaluate_potential(self, system):
        """Evaluate the potential energy.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential in.

        Returns
        -------
        out : float
            The potential energy.

        """
        v_pot = None
        for pot in self.potential:
            if v_pot is None:
                v_pot = pot.potential(system)
            else:
                v_pot += pot.potential(system)
        return v_pot

    def evaluate_potential_and_force(self, system):
        """Evaluate the potential energy and the force.

        Parameters
        ----------
        system : object like :py:class:`.System`
            The system we evaluate the potential energy and force in.

        Returns
        -------
        out[0] : float
            The potential energy.
        out[1] : numpy.array
            The calculated forces.
        out[2] : numpy.array
            The calculated virial.

        """
        v_pot = None
        force = None
        virial = None
        for pot in self.potential:
            if v_pot is None or force is None or virial is None:
                v_pot, force, virial = pot.potential_and_force(system)
            else:
                v_poti, forcei, viriali = pot.potential_and_force(system)
                v_pot += v_poti
                force += forcei
                virial += viriali
        return v_pot, force, virial

    def __str__(self):
        """Return a string representation of the force field.

        The string representation is built using the string
        descriptions of the potential functions.

        Returns
        -------
        out : string
            Description of the force field and the potential functions
            included in the force field.

        """
        msg = [f'Force field: {self.desc}']
        if len(self.potential) < 1:
            msg.append('No potential functions added yet!')
        else:
            msg.append('Potential functions:')
            for i, pot in enumerate(self.potential):
                msg.append(f'{i + 1}: {pot}')
        return '\n'.join(msg)

    def print_potentials(self):
        """Print information on potentials in the force field.

        This is intended as a lighter alternative to `self.__str__`
        which can be verbose. This function will not actually do the
        printing, but it returns a string which can be printed.

        Returns
        -------
        out : string
            Description of the potential functions in this force field.

        """
        msg = [f'Force field: {self.desc}']
        for i, pot in enumerate(self.potential):
            msg.append(f'\t{i + 1}: {pot.desc}')
        return '\n'.join(msg)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test the methods in pyretis.setup.createsimulation"""
import os
import logging
import unittest
import numpy as np
from io import StringIO
from unittest.mock import patch
from pyretis.engines.internal import VelocityVerlet
from pyretis.core.units import create_conversion_factors
from pyretis.core.system import System
from pyretis.core.particles import Particles
from pyretis.core.pathensemble import (
    PathEnsemble,
    PathEnsembleExt,
)
from pyretis.setup.createsimulation import (
    create_ensemble,
    create_ensembles,
    create_nve_simulation,
    create_mdflux_simulation,
    create_md_simulation,
    create_umbrellaw_simulation,
    create_tis_simulation,
    create_retis_simulation,
    create_simulation,
)
from pyretis.tools.lattice import generate_lattice
from pyretis.forcefield.forcefield import ForceField
from pyretis.forcefield.potentials import PairLennardJonesCutnp
from pyretis.core.box import create_box
from pyretis.inout.settings import SECTIONS

from pyretis.simulation.simulation import Simulation
from pyretis.simulation.md_simulation import (
    SimulationNVE,
    SimulationMD,
    SimulationMDFlux,
)
from pyretis.simulation.mc_simulation import UmbrellaWindowSimulation
from pyretis.simulation.path_simulation import (
    SimulationTIS,
    SimulationRETIS
)
from pyretis.inout.settings import fill_up_tis_and_retis_settings
from pyretis.inout.checker import check_ensemble

logging.disable(logging.CRITICAL)


HERE = os.path.abspath(os.path.dirname(__file__))

ORDER_ERROR_MSG = (
    'No order parameter was defined,'
    ' but the engine *does* require it.'
)
MISSING_STEPS_ERROR_MSG = 'Simulation setting "steps" is missing!'
MISSING_INTERFACES_ERROR_MSG = 'Simulation setting "interfaces" is missing!'


def create_test_settings():
    """Create a settings we can use for testing."""
    settings = {'simulation': {'task': 'retis',
                               'interfaces': [-1., 0, 1],
                               'steps': 10,
                               'exe_path': HERE,
                               'zero_ensemble': True,
                               'flux': True},
                'orderparameter': {'class': 'Position',
                                   'dim': 'x',
                                   'index': 0,
                                   'periodic': False},
                'particles': {'type': 'internal'},
                'system': {'units': 'lj',
                           'obj': create_test_system()},
                'engine': {'obj': VelocityVerlet(0.002)},
                'tis': {'shooting_moves': ['sh', 'sh', 'sh']}}
    return settings


def create_test_system():
    """Create a system we can use for testing."""
    create_conversion_factors('lj')
    xyz, size = generate_lattice('fcc', [2, 2, 2], density=0.9)
    size = np.array(size)
    box = create_box(low=size[:, 0], high=size[:, 1])
    system = System(units='lj', box=box, temperature=2.0)
    system.particles = Particles(dim=3)
    for pos in xyz:
        system.add_particle(pos, vel=np.zeros_like(pos),
                            force=np.zeros_like(pos),
                            mass=1.0, name='Ar', ptype=0)
    gen_settings = {'distribution': 'maxwell', 'momentum': True, 'seed': 0}
    system.generate_velocities(**gen_settings)
    potentials = [
        PairLennardJonesCutnp(dim=3, shift=True, mixing='geometric'),
    ]
    parameters = [
        {0: {'sigma': 1, 'epsilon': 1, 'rcut': 2.5}},
    ]
    system.forcefield = ForceField(
        'Lennard Jones force field',
        potential=potentials,
        params=parameters,
    )
    return system


class TestMethods(unittest.TestCase):
    """Test some of the methods from .createsimulation."""
    def test_create_path_ensemble(self):
        """Test create_path_ensemble."""
        path_ensemble = PathEnsemble(ensemble_number=2, interfaces=[-1., 0, 1])
        self.assertEqual(path_ensemble.ensemble_number, 2)
        self.assertIsInstance(path_ensemble, PathEnsemble)
        path_ensemble = PathEnsembleExt(ensemble_number=2,
                                        interfaces=[-1., 0, 1])
        self.assertIsInstance(path_ensemble, PathEnsembleExt)
        self.assertEqual(path_ensemble.ensemble_number, 2)
        # Test with some "missing" settings:
        settings = {'simulation': {'interfaces': [-1., 0.]}}
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(KeyError):
                create_ensemble(settings)

    def test_create_ensemble(self):
        """Test create_ensemble."""
        settings = create_test_settings()

        fill_up_tis_and_retis_settings(settings)
        check_ensemble(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        for i_ens, ensemble in enumerate(ensembles):
            path_ensemble = ensemble['path_ensemble']
            self.assertEqual(path_ensemble.ensemble_number, i_ens)
            self.assertIsInstance(path_ensemble, PathEnsemble)

        for i_ens in range(len(ensembles)):
            settings['ensemble'][i_ens]['particles']['type'] = 'external'

        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        for i_ens, ensemble in enumerate(ensembles):
            path_ensemble = ensemble['path_ensemble']
            self.assertIsInstance(path_ensemble, PathEnsembleExt)
            self.assertEqual(path_ensemble.ensemble_number, i_ens)

    def test_create_ensemble_permeability(self):
        """Test create_ensemble."""
        settings = create_test_settings()
        settings['simulation'] = {'task': 'retis',
                                  'interfaces': [-1., 0, 1],
                                  'exe_path': HERE,
                                  'zero_ensemble': True,
                                  'flux': True,
                                  'permeability': True,
                                  'zero_left': -2}

        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        ens = ensembles[0]
        self.assertEqual((-2, -1.5, -1), ens['path_ensemble'].interfaces)
        self.assertEqual(["R", "L"], ens['path_ensemble'].start_condition)

    def test_create_ensemble_pptis(self):
        """Test create_ensemble for pptis."""
        # with flux=False, zero_ensemble=True
        # -------------------------------------
        settings = create_test_settings()
        settings['tis']['shooting_moves'] += ['sh', 'sh']
        settings['simulation'] = {'task': 'pptis',
                                  'interfaces': [-1., 0., 1., 2., 3.],
                                  'flux': False,
                                  'zero_ensemble': True,
                                  'exe_path': HERE}

        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        n_ens = len(ensembles)
        for i in range(n_ens):
            self.assertEqual(ensembles[i]['path_ensemble'].start_condition,
                             ['R', 'L'])
        hoped = [(-1., -1., 0.), (-1., 0., 1.),
                 (0., 1., 2.), (1., 2., 3.)]
        self.assertEqual(len(hoped), n_ens)
        for ensemble, hope in zip(ensembles, hoped):
            self.assertEqual(ensemble['path_ensemble'].interfaces, hope)

    def test_create_ensemble_pptis_1(self):
        """Test create_ensemble for pptis."""
        # with flux=True, zero_ensemble=True
        # -------------------------------------
        settings = create_test_settings()
        settings['tis']['shooting_moves'] += ['sh', 'sh']
        settings['simulation'] = {'task': 'pptis',
                                  'interfaces': [-1., 0., 1., 2., 3.],
                                  'flux': True,
                                  'zero_ensemble': True,
                                  'exe_path': HERE}

        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        n_ens = len(ensembles)
        # start condition
        self.assertEqual(ensembles[0]['path_ensemble'].start_condition, 'R')
        for i in range(1, n_ens):
            self.assertEqual(ensembles[i]['path_ensemble'].start_condition,
                             ['R', 'L'])
        # interfaces
        hoped = [(-np.inf, -1., -1.), (-1., -1., 0.), (-1., 0., 1.),
                 (0., 1., 2.), (1., 2., 3.)]
        self.assertEqual(len(hoped), n_ens)
        for ensemble, hope in zip(ensembles, hoped):
            self.assertEqual(ensemble['path_ensemble'].interfaces, hope)

    def test_create_ensemble_pptis_2(self):
        """Test create_ensemble for repptis."""
        # with flux=False, zero_ensemble=False
        # -------------------------------------
        settings = create_test_settings()
        settings['tis']['shooting_moves'] += ['sh', 'sh']
        settings['simulation'] = {'task': 'pptis',
                                  'interfaces': [-1., 0., 1., 2., 3.],
                                  'flux': False,
                                  'zero_ensemble': False,
                                  'exe_path': HERE}
        # just [1+], [2+], [3+]
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        n_ens = len(ensembles)
        # start condition
        for i in range(n_ens):
            self.assertEqual(ensembles[i]['path_ensemble'].start_condition,
                             ['R', 'L'])
        # interfaces
        hoped = [(-1., 0., 1.), (0., 1., 2.), (1., 2., 3.)]
        self.assertEqual(len(hoped), n_ens)
        for ensemble, hope in zip(ensembles, hoped):
            self.assertEqual(ensemble['path_ensemble'].interfaces, hope)

    def test_create_ensemble_pptis_3(self):
        """Test create_ensemble for pptis."""
        # with flux=True, zero_ensemble=False
        # ------------------------------------
        settings = create_test_settings()
        settings['tis']['shooting_moves'] += ['sh', 'sh']
        settings['simulation'] = {'task': 'pptis',
                                  'interfaces': [-1., 0., 1., 2., 3.],
                                  'flux': True,
                                  'zero_ensemble': False,
                                  'exe_path': HERE}
        # [0-], [1+], [2+], [3+]
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        n_ens = len(ensembles)
        # start condition
        self.assertEqual(ensembles[0]['path_ensemble'].start_condition, 'R')
        for i in range(1, n_ens):
            self.assertEqual(ensembles[i]['path_ensemble'].start_condition,
                             ['R', 'L'])
        # interfaces
        hoped = [(-np.inf, -1., -1.), (-1., 0., 1.),
                 (0., 1., 2.), (1., 2., 3.)]
        self.assertEqual(len(hoped), n_ens)
        for ensemble, hope in zip(ensembles, hoped):
            self.assertEqual(ensemble['path_ensemble'].interfaces, hope)

    def test_create_ensemble_repptis_0(self):
        """Test create_ensemble for repptis."""
        # with flux=False, zero_ensemble=False, no permeability
        settings = create_test_settings()
        settings['simulation'] = {'task': 'repptis',
                                  'interfaces': [-1., 0., 1., 2., 3.],
                                  'flux': False,
                                  'zero_ensemble': False,
                                  'exe_path': HERE}
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        n_ens = len(ensembles)
        for i in range(n_ens):
            self.assertEqual(ensembles[i]['path_ensemble'].start_condition,
                             ['R', 'L'])
        hoped = [(-1., 0., 1.), (0., 1., 2.), (1., 2., 3.)]
        self.assertEqual(len(hoped), n_ens)
        for ensemble, hope in zip(ensembles, hoped):
            self.assertEqual(ensemble['path_ensemble'].interfaces, hope)

    def test_create_ensemble_repptis_1(self):
        """Test create_ensemble for repptis."""
        # with flux=False, zero_ensemble=False, with permeability
        settings = create_test_settings()
        settings['simulation'] = {'task': 'repptis',
                                  'interfaces': [-1., 0., 1., 2., 3.],
                                  'exe_path': HERE,
                                  'zero_ensemble': True,
                                  'flux': True,
                                  'zero_left': -2.,
                                  'permeability': True}
        # [0-], [0+], [1+], [2+], [3+]
        settings['tis']['shooting_moves'] += ['sh', 'sh']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        # start_condition for paths
        n_ens = len(ensembles)
        self.assertEqual(ensembles[0]['path_ensemble'].start_condition,
                         ['R', 'L'])
        for i in range(1, n_ens):
            self.assertEqual(ensembles[i]['path_ensemble'].start_condition,
                             ['R', 'L'])
        hoped = [(-2, -1.5, -1), (-1., -1., 0.), (-1., 0., 1.),
                 (0., 1., 2.), (1., 2., 3.)]
        self.assertEqual(len(hoped), n_ens)
        for ensemble, hope in zip(ensembles, hoped):
            self.assertEqual(ensemble['path_ensemble'].interfaces, hope)

    def test_create_ensemble_repptis_2(self):
        """Test create_ensemble for repptis."""
        settings = create_test_settings()
        settings['simulation'] = {'task': 'repptis',
                                  'interfaces': [-1., 0., 1., 2., 3.],
                                  'exe_path': HERE,
                                  'flux': False,
                                  'zero_ensemble': True,
                                  }
        settings['tis']['shooting_moves'] += ['sh']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        self.assertEqual(ensembles[-1]['path_ensemble'].start_condition,
                         ['R', 'L'])
        hoped = [(-1., -1., 0.), (-1., 0., 1.),
                 (0., 1., 2.), (1., 2., 3.)]
        self.assertEqual(len(hoped), len(ensembles))
        for ensemble, hope in zip(ensembles, hoped):
            self.assertEqual(ensemble['path_ensemble'].interfaces, hope)

    def test_raise_on_wrong_mirror(self):
        """Test bad mirror."""
        settings = create_test_settings()
        settings['simulation'] = {'task': 'retis',
                                  'interfaces': [-1., 0, 1],
                                  'exe_path': HERE,
                                  'zero_ensemble': True,
                                  'flux': True,
                                  'zero_left': -2,
                                  'permeability': True}
        settings['orderparameter'] = {'class': 'Permeability',
                                      'dim': 'x',
                                      'index': 0,
                                      'mirror_pos': 0}

        fill_up_tis_and_retis_settings(settings)
        with self.assertRaisesRegex(ValueError, "have a mirror at -1.5"):
            with patch('sys.stdout', new=StringIO()):
                create_ensembles(settings)

    def test_create_ensembles(self):
        """Test create_ensembles."""
        settings = create_test_settings()
        settings['simulation'] = {'task': 'tis',
                                  'interfaces': [-1., 0., 1.0, 2.0],
                                  'exe_path': HERE,
                                  'zero_ensemble': False,
                                  'flux': False}
        fill_up_tis_and_retis_settings(settings)
        check_ensemble(settings)
        with patch('sys.stdout', new=StringIO()):
            ensembles = create_ensembles(settings)
        detects = [1.0, 2.0]
        self.assertEqual(len(ensembles), 2)
        for iens, ens in enumerate(ensembles):
            self.assertIsInstance(ens['path_ensemble'], PathEnsemble)
        for iens, ens in enumerate(settings['ensemble']):
            self.assertEqual(detects[iens], ens['tis']['detect'])

    def test_fail_simulation(self):
        """Test fail simulation."""
        settings = {'simulation': {'task': 'fake'}}
        with self.assertRaises(ValueError) as err:
            create_simulation(settings)
        self.assertEqual('Unknown simulation task fake',
                         str(err.exception))

    def test_load_simulation(self):
        """Test cycle counter."""
        settings = {}
        sim = Simulation(settings, controls={'steps': 16})
        restart = {'simulation': {'cycle': {'steps': 30,
                                            'endcycle': 30}}}
        sim.load_restart_info(restart)
        self.assertTrue(sim.cycle['steps'] == 16)
        self.assertTrue(sim.cycle['step'] == 0)

    def test_create_nve_simulation(self):
        """Test create_nve_simulation."""
        settings = {'simulation': {'task': 'md-nve', 'steps': 10},
                    'system': {'obj': create_test_system()},
                    'engine': {'obj': VelocityVerlet(0.002)}}
        sim = create_nve_simulation(settings)
        self.assertIsInstance(sim, SimulationNVE)
        self.assertEqual(sim.simulation_type,
                         SimulationNVE.simulation_type)
        del settings['simulation']['steps']
        with self.assertRaises(ValueError) as err:
            create_nve_simulation(settings)
        self.assertEqual(MISSING_STEPS_ERROR_MSG, str(err.exception))

    def test_create_md_simulation(self):
        """Test create_nve_simulation."""
        settings = {'simulation': {'task': 'md', 'steps': 10},
                    'system': {'obj': create_test_system()},
                    'engine': {'obj': VelocityVerlet(0.002)}}
        sim = create_md_simulation(settings)
        self.assertIsInstance(sim, SimulationMD)
        self.assertEqual(sim.simulation_type,
                         SimulationMD.simulation_type)
        del settings['simulation']['steps']
        with self.assertRaises(ValueError) as err:
            create_md_simulation(settings)
        self.assertEqual(MISSING_STEPS_ERROR_MSG, str(err.exception))

    def test_create_mdflux_simulation(self):
        """Test create_mdflux_simulation."""
        settings = {'simulation': {'task': 'mdflux'},
                    'system': {'obj': create_test_system()},
                    'engine': {'obj': VelocityVerlet(0.002)}}

        with self.assertRaises(ValueError) as err:
            create_mdflux_simulation(settings)
        self.assertEqual(ORDER_ERROR_MSG, str(err.exception))
        settings['orderparameter'] = {'class': 'Position',
                                      'dim': 'x', 'index': 0,
                                      'periodic': False}
        # Test that we fail because required setting "steps" are missing:
        with self.assertRaises(ValueError) as err:
            create_mdflux_simulation(settings)
        self.assertEqual(MISSING_STEPS_ERROR_MSG, str(err.exception))
        settings['simulation']['steps'] = 10
        # Test that we fail because required setting "interfaces" are
        # missing:
        with self.assertRaises(ValueError) as err:
            create_mdflux_simulation(settings)
        self.assertEqual(MISSING_INTERFACES_ERROR_MSG, str(err.exception))
        settings['simulation']['interfaces'] = [0, 1]
        sim = create_mdflux_simulation(settings)
        self.assertIsInstance(sim, SimulationMDFlux)

    def test_create_umbrellawsimulation(self):
        """Test create_umbrellaw_simulation."""
        settings = {'simulation': {'task': 'umbrellawindow', 'seed': 3},
                    'system': {'obj': create_test_system}}

        with self.assertRaises(ValueError) as err:
            create_umbrellaw_simulation(settings)
        self.assertEqual('Simulation setting "umbrella" is missing!',
                         str(err.exception))

        settings['simulation']['umbrella'] = [-1.0, 0.0]
        with self.assertRaises(ValueError) as err:
            create_umbrellaw_simulation(settings)
        self.assertEqual('Simulation setting "overlap" is missing!',
                         str(err.exception))

        settings['simulation']['overlap'] = -0.5
        with self.assertRaises(ValueError) as err:
            create_umbrellaw_simulation(settings)
        self.assertEqual('Simulation setting "maxdx" is missing!',
                         str(err.exception))

        settings['simulation']['maxdx'] = 1.0
        with self.assertRaises(ValueError) as err:
            create_umbrellaw_simulation(settings)
        self.assertEqual('Simulation setting "mincycle" is missing!',
                         str(err.exception))
        settings['simulation']['mincycle'] = 10
        sim = create_umbrellaw_simulation(settings)
        self.assertIsInstance(sim, UmbrellaWindowSimulation)

        del settings['system']
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(KeyError) as err:
                create_umbrellaw_simulation(settings)
        self.assertEqual("'system'", str(err.exception))

    def test_create_tis_simulation(self):
        """Test create_tis_simulation."""
        settings = {'simulation': {'task': 'tis',
                                   'interfaces': [-1., 0., 1.],
                                   'exe_path': HERE,
                                   'zero_ensemble': False,
                                   'flux': False},
                    'tis': SECTIONS['tis'],
                    'particles': {'type': 'internal'},
                    'system': {'obj': create_test_system()},
                    'engine': {'obj': VelocityVerlet(0.002)},
                    'output': SECTIONS['output']}
        fill_up_tis_and_retis_settings(settings)

        # Test that the set up fails, when we did not define an
        # order parameter and the engine wants it:
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                create_tis_simulation(settings)
        self.assertEqual(ORDER_ERROR_MSG, str(err.exception))
        # Continue testing with an order parameter defined:
        settings['orderparameter'] = {'class': 'Position',
                                      'dim': 'x', 'index': 0,
                                      'periodic': False}
        # Test that we fail when we are missing some settings:
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                create_tis_simulation(settings)
        # it should raise also missing steps
        self.assertEqual(MISSING_STEPS_ERROR_MSG, str(err.exception))
        settings['simulation']['steps'] = 10
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            sim = create_tis_simulation(settings)
        self.assertIsInstance(sim, SimulationTIS)
        del settings['ensemble']
        del settings['system']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(KeyError) as err:
                create_tis_simulation(settings)
        self.assertEqual("'system'", str(err.exception))

    def test_create_tis_simulations(self):
        """Test create_tis_simulations."""
        settings = create_test_settings()
        settings['simulation'] = {'task': 'tis',
                                  'interfaces': [-1., 0., 1., 2],
                                  'steps': 10,
                                  'exe_path': HERE,
                                  'zero_ensemble': False,
                                  'flux': False}
        settings['tis'] = SECTIONS['tis']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            simulations = create_tis_simulation(settings)
        self.assertIsInstance(simulations, SimulationTIS)
        del settings['ensemble']
        del settings['engine']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                create_tis_simulation(settings)
        self.assertEqual("Could not create engine from settings!",
                         str(err.exception))

    def test_create_retis_simulation(self):
        """Test create_retis_simulation."""
        settings = {'simulation': {'task': 'retis',
                                   'interfaces': [-1., 0., 1.],
                                   'exe_path': '.',
                                   'zero_ensemble': True,
                                   'flux': True},
                    'tis': SECTIONS['tis'],
                    'retis': SECTIONS['retis'],
                    'engine': {'obj': VelocityVerlet(0.002)},
                    'particles': {'type': 'internal'},
                    'system': {'type': 'internal',
                               'units': 'lj',
                               'obj': create_test_system()}}

        del settings['tis']['ensemble_number']
        fill_up_tis_and_retis_settings(settings)

        # Test that we fail without an order parameter defined:
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                create_retis_simulation(settings)
        self.assertEqual(ORDER_ERROR_MSG, str(err.exception))
        settings['orderparameter'] = {'class': 'Position',
                                      'dim': 'x', 'index': 0,
                                      'periodic': False}

        fill_up_tis_and_retis_settings(settings)
        # Test that we fail when we are missing some settings:
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                create_retis_simulation(settings)
        self.assertEqual(MISSING_STEPS_ERROR_MSG, str(err.exception))
        settings['simulation']['steps'] = 10
        with patch('sys.stdout', new=StringIO()):
            sim = create_retis_simulation(settings)
        self.assertIsInstance(sim, SimulationRETIS)
        del settings['ensemble']
        del settings['engine']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(ValueError) as err:
                create_tis_simulation(settings)
        self.assertEqual("Could not create engine from settings!",
                         str(err.exception))

    def test_create_retis_simulation2(self):
        """Test create_simulation for SimulationRETIS."""
        settings = create_test_settings()
        settings['tis'] = SECTIONS['tis']
        settings['retis'] = SECTIONS['retis']

        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            sim = create_retis_simulation(settings)
        self.assertIsInstance(sim, SimulationRETIS)
        del settings['ensemble']
        del settings['system']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            with self.assertRaises(KeyError) as err:
                create_tis_simulation(settings)
        self.assertEqual("'system'", str(err.exception))

    def test_create_simulation(self):
        """Test create_simulation."""
        settings = {'simulation': {'task': 'does-not-exist'}}
        with self.assertRaises(ValueError) as err:
            create_simulation(settings)
        self.assertEqual('Unknown simulation task does-not-exist',
                         str(err.exception))

    def test_create_simulationnve(self):
        """Test create_simulation for NVE"""
        settings = {'simulation': {'steps': 10, 'task': 'md-nve'},
                    'system': {'obj': create_test_system()},
                    'engine': {'obj': VelocityVerlet(0.002)}}
        sim = create_simulation(settings)
        self.assertIsInstance(sim, SimulationNVE)

    def test_create_simulationmdflux(self):
        """Test create_simulation for MD-Flux"""
        settings = create_test_settings()
        settings['simulation'] = {'steps': 10,
                                  'task': 'md-flux',
                                  'interfaces': [-1.0]}
        sim = create_mdflux_simulation(settings)
        self.assertIsInstance(sim, SimulationMDFlux)

    def test_create_simulationumb(self):
        """Test create_simulation for UmbrellaWindow"""
        settings = create_test_settings()
        settings['simulation'] = {'task': 'umbrellawindow',
                                  'umbrella': [-1.0, 0.0],
                                  'overlap': -0.5,
                                  'maxdx': 1.0,
                                  'mincycle': 10}
        sim = create_umbrellaw_simulation(settings)
        self.assertIsInstance(sim, UmbrellaWindowSimulation)

    def test_create_simulationtis2(self):
        """Test create_simulation for SimulationTIS multiple."""
        # TIS-multiple:
        settings = create_test_settings()
        settings['simulation'] = {'task': 'tis',
                                  'interfaces': [-1., 0., 1., 2.0],
                                  'steps': 10,
                                  'exe_path': HERE,
                                  'zero_ensemble': False,
                                  'flux': False}
        settings['tis'] = SECTIONS['tis']
        fill_up_tis_and_retis_settings(settings)
        with patch('sys.stdout', new=StringIO()):
            simulations = create_tis_simulation(settings)
        self.assertIsInstance(simulations, SimulationTIS)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Simple script to compare the outcome of two simulations.

The outcome of the ``md_nve.py`` simulation should be independent (to numerical
precision) of the units used. This script will test that by comparing:

- the output in `thermo.txt`

- the generated trajectory, `traj.xyz`

For the energies, it will create a plot comparing the energies, the pressure
and the temperature.
"""
import sys
import colorama
import numpy as np
from matplotlib import pyplot as plt
from matplotlib import gridspec
from pyretis.inout import print_to_screen
from pyretis.inout.formats.snapshot import SnapshotFile
from pyretis.inout.settings import parse_settings_file
from pyretis.core.units import (
    create_conversion_factors,
    generate_system_conversions,
    CONVERT
)
plt.style.use('seaborn-colorblind')


def snapshot_difference(snap1, snap2, unit1, unit2):
    """Calculate difference between two snapshots."""
    xyz1 = np.column_stack((snap1['x'], snap1['y'], snap1['z']))
    xyz2 = np.column_stack((snap2['x'], snap2['y'], snap2['z']))
    xyz2 *= CONVERT['length'][unit1, unit2]
    diff = (xyz1 - xyz2)**2
    dsum = np.einsum('ij,ij -> i', diff, diff)
    vel1 = np.column_stack((snap1['vx'], snap1['vy'], snap1['vz']))
    vel2 = np.column_stack((snap2['vx'], snap2['vy'], snap2['vz']))
    vel2 *= CONVERT['velocity'][unit1, unit2]
    diffv = (vel1 - vel2)**2
    dsumv = np.einsum('ij,ij -> i', diffv, diffv)
    return sum(dsum), sum(dsumv)


def compare_traj(traj1, traj2, unit1, unit2, tol=1e-12):
    """Compare two trajectories from PyRETIS.

    Here we calculate the mean squared error for the two
    trajectories.

    Parameters
    ----------
    traj1 : string
        A trajectory file to open.
    traj2 : string
        A trajectory file to open.
    unit1 : string
        The system of units for the first trajectory.
    unit2 : string
        The system of units for the second trajectory.
    tol : float
        A tolerance for comparing numbers.

    Returns
    -------
    None, just prints out the result of the comparison.

    """
    retval = 1
    print_to_screen('Comparing trajectories', level='info')
    print('Checking mean squared error...')
    file1 = SnapshotFile(traj1, 'r').load()
    file2 = SnapshotFile(traj2, 'r').load()
    error, error_v = 0.0, 0.0
    nsnap = 0
    for snap1, snap2 in zip(file1, file2):
        pose, vele = snapshot_difference(snap1, snap2, unit1, unit2)
        error += pose
        error_v += vele
        nsnap += 1
    error /= float(nsnap)
    if abs(error) < tol:
        lev = 'success'
        retval = 0
    else:
        lev = 'error'
        return 1
    print_to_screen(f'Mean error - positions: {error}',
                    level=lev)
    if abs(error_v) < tol:
        lev = 'success'
        retval = 0
    else:
        lev = 'error'
        return 1
    print_to_screen(f'Mean error - velocities: {error_v}',
                    level=lev)
    return retval


def run_comparison():
    """Run the comparison."""
    settings = parse_settings_file('settings.rst')
    unit = settings['system']['units']
    create_conversion_factors('lj')
    create_conversion_factors(unit)
    generate_system_conversions('lj', unit)
    ret = compare_traj('../traj.xyz', 'traj.xyz', unit, 'lj')
    return unit, 'lj', ret


def make_plot(unit1, unit2):
    """Plot for comparison."""
    ljunits = np.loadtxt('../thermo.txt')
    other_units = np.loadtxt('thermo.txt')
    # convert other_units:
    other_units[:, 1] *= CONVERT['temperature'][unit1, unit2]
    other_units[:, 2:5] *= CONVERT['energy'][unit1, unit2]
    other_units[:, 5] *= CONVERT['pressure'][unit1, unit2]
    fig1 = plt.figure(figsize=(12, 6))
    grid = gridspec.GridSpec(2, 2)
    ax1 = fig1.add_subplot(grid[:, 0])
    ljlab = f'"{unit2}"'
    unilab = f'"{unit1}"'
    ax1.plot([], [], label='Potential', lw=0, alpha=0)
    ax1.plot([], [], label='Kinetic', lw=0, alpha=0)
    ax1.plot([], [], label='Total', lw=0, alpha=0)
    ax1.plot(ljunits[:, 0], ljunits[:, 2], label=ljlab,
             ls='-', lw=7, alpha=0.8)
    ax1.plot(ljunits[:, 0], ljunits[:, 3], label=ljlab,
             ls='-', lw=7, alpha=0.8)
    ax1.plot(ljunits[:, 0], ljunits[:, 4], label=ljlab,
             ls='-', lw=7, alpha=0.8)
    ax1.plot(other_units[:, 0], other_units[:, 2], label=unilab,
             ls='--', lw=3, alpha=0.8)
    ax1.plot(other_units[:, 0], other_units[:, 3], label=unilab,
             ls='--', lw=3, alpha=0.8)
    ax1.plot(other_units[:, 0], other_units[:, 4], label=unilab,
             ls='--', lw=3, alpha=0.8)
    ax1.set_xlabel('Step no.')
    ax1.set_ylabel('Energy per particle')
    _ = ax1.legend(loc='center left', prop={'size': 'large'}, ncol=3)

    ax2 = fig1.add_subplot(grid[0, 1])
    ax2.plot(ljunits[:, 0], ljunits[:, 1], label=ljlab,
             ls='-', lw=7, alpha=0.9)
    ax2.plot(other_units[:, 0], other_units[:, 1], label=unilab,
             ls='--', lw=3, alpha=0.8, color='0.2')
    ax2.set_ylabel('Temperature')
    ax2.legend(loc='upper right', prop={'size': 'large'})
    ax3 = fig1.add_subplot(grid[1, 1])
    ax3.plot(ljunits[:, 0], ljunits[:, 5], label=ljlab,
             ls='-', lw=7, alpha=0.9)
    ax3.plot(other_units[:, 0], other_units[:, 5], label=unilab,
             ls='--', lw=3, alpha=0.8, color='0.2')
    ax3.set_xlabel('Step no.')
    ax3.set_ylabel('Pressure')
    ax3.legend(loc='lower right', prop={'size': 'large'})
    fig1.subplots_adjust(bottom=0.12, right=0.95, top=0.95,
                         left=0.08, wspace=0.2)
    fig1.savefig('compare.png')


def main(args):
    """Run the comparison."""
    uni1, uni2, ret = run_comparison()
    if 'plot' in args:
        make_plot(uni1, uni2)
        plt.show()
    return ret


if __name__ == '__main__':
    colorama.init(autoreset=True)
    sys.exit(main(sys.argv[1:]))



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test block error analysis."""
import unittest
import numpy as np
from pyretis.analysis import block_error
from pyretis.inout.analysisio.analysisio import analyse_file


class TestAnalysis(unittest.TestCase):
    """Test that we can do block error analysis."""

    def test_block_error_weights(self):
        data = np.random.random(100)
        weights = np.ones(data.shape)*(np.random.random() + 0.001)  # catch 0.0
        r1 = block_error(data)
        r2 = block_error(data, weights=weights)
        r3 = block_error(np.stack([data, weights], axis=-1))
        for i, j, k in zip(r1, r2, r3):
            self.assertTrue(np.allclose(i, j))
            self.assertTrue(np.allclose(i, k))

    def test_block_error_zero_weights(self):
        # This has to be even to not alter the number of blocks
        data = np.random.random(100)
        weights = np.ones(data.shape)

        # add a giant data point that should not matter
        data2 = np.array([i for i in data]+[1e10])
        weights2 = np.array([i for i in weights]+[0])
        r1 = block_error(data, weights=weights)
        r2 = block_error(data2, weights=weights2)
        for i, j in zip(r1, r2):
            self.assertTrue(np.allclose(i, j))

    def test_block_error_zero_weights_start(self):
        data = np.random.random(100)
        weights = np.ones(data.shape)
        weights[0] = 0  # this would lead to nan's
        r1 = block_error(data, weights=weights)
        self.assertFalse(np.isnan(r1[1][0]))

    def test_analyse_file_fail(self):
        """Test the function to fail."""
        with self.assertRaises(ValueError) as err:
            analyse_file('no_ext', 'non_existent', {})
        self.assertIn('Unknown file type', str(err.exception))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test for the simple potentials."""
import logging
import unittest
import numpy as np
from pyretis.core.system import System
from pyretis.core.particles import Particles
from pyretis.forcefield.potentials import DoubleWell, RectangularWell
logging.disable(logging.CRITICAL)


def evaluate_potential(positions, system, function):
    """Just evaluate the given potential"""
    all_pot, all_pot2 = [], []
    all_force, all_force2 = [], []
    all_vir, all_vir2 = [], []
    for i in positions:
        system.particles.pos = i
        pot = function.potential(system)
        all_pot.append(pot)
        force, vir = function.force(system)
        all_force.append(force)
        all_vir.append(vir)
        pot2, force2, vir2 = function.potential_and_force(system)
        all_pot2.append(pot2)
        all_force2.append(force2)
        all_vir2.append(vir2)
    out = []
    for i in (all_pot, all_pot2, all_force, all_force2, all_vir, all_vir2):
        out.append(np.array(i))
    return out


class TestDoubleWell(unittest.TestCase):
    """Test the double well potential."""

    def test_doublewell_potential(self):
        """Test the double well potential."""
        params = {'a': 1.0, 'b': 2.0, 'c': 3.0}
        function = DoubleWell(**params)
        system = System()
        system.particles = Particles(dim=system.get_dim())
        system.add_particle(0.0)
        pos2 = np.linspace(-10, 10, 100)
        pos = np.reshape(np.linspace(-10, 10, 100), (100, 1, 1))
        out = evaluate_potential(pos, system, function)
        all_pot2 = (params['a'] * pos2**4 -
                    params['b'] * (pos2 - params['c'])**2)
        self.assertTrue(np.allclose(out[0], out[1]))
        self.assertTrue(np.allclose(out[2], out[3]))
        self.assertTrue(np.allclose(out[4], out[5]))
        self.assertTrue(np.allclose(out[0], all_pot2))
        all_force2 = (-4.0 * params['a'] * pos**3 +
                      2.0 * params['b'] * (pos - params['c']))
        self.assertTrue(np.allclose(out[2], all_force2))

    def test_more_particles(self):
        """Test what happens if we have more particles."""
        params = {'a': 1.0, 'b': 2.0, 'c': 3.0}
        npart = 10
        function = DoubleWell(**params)
        system = System()
        system.particles = Particles(dim=system.get_dim())
        for _ in range(npart):
            system.add_particle(0.0)
        pos = np.linspace(-10, 10, 100)
        all_pot = []
        for i in pos:
            for j in range(npart):
                system.particles.pos[j][0] = i
            pot = function.potential(system)
            all_pot.append(pot)
        all_pot = np.array(all_pot)
        all_pot2 = params['a'] * pos**4 - params['b'] * (pos - params['c'])**2
        self.assertTrue(np.allclose(all_pot, all_pot2 * npart))


class TestRectangularWell(unittest.TestCase):
    """Test the rectangular well potential."""

    def test_initiation(self):
        """Test the way we give input."""
        params = {'left': -1.0, 'right': 1.0}
        RectangularWell(**params)
        params = {'left': 1.0, 'right': -1.0}
        logging.disable(logging.INFO)
        with self.assertLogs('pyretis.forcefield.potentials.potentials',
                             level='WARNING'):
            RectangularWell(**params)
        logging.disable(logging.CRITICAL)

    def test_rectwell_potential(self):
        """Test the double well potential."""
        params = {'left': -1.0, 'right': 1.0}
        function = RectangularWell(**params)
        system = System()
        system.particles = Particles(dim=system.get_dim())
        system.add_particle(0.0)
        pos = np.linspace(-10, 10, 100)
        for i in pos:
            system.particles.pos[0][0] = i
            pot = function.potential(system)
            if params['left'] < i < params['right']:
                self.assertEqual(pot, 0.0)
            else:
                self.assertEqual(pot, float('inf'))
        for i in (params['left'], params['right']):
            system.particles.pos[0][0] = i
            pot = function.potential(system)
            self.assertEqual(pot, float('inf'))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Example of running a MD simulation using the PyRETIS library.

The system considered is a simple Lennard-Jones fluid.

"""
import sys
import numpy as np
from matplotlib import pyplot as plt
from pyretis.core import create_box, Particles, System
from pyretis.simulation import SimulationNVE
from pyretis.engines import VelocityVerlet
from pyretis.tools import generate_lattice
from pyretis.forcefield import ForceField
from pyretis.forcefield.potentials import PairLennardJonesCutnp


# If you have a recent version of matplotlib you can make
# the plot nicer by loading a style, e.g.:
plt.style.use('seaborn-colorblind')


def set_up_system():
    """Set up a system using the PyRETIS library."""
    print('Creating box:')
    xyz, size = generate_lattice('fcc', [3, 3, 3], density=0.9)
    size = np.array(size)
    box = create_box(low=size[:, 0], high=size[:, 1])
    print(box)

    print('Creating system:')
    system = System(units='lj', box=box, temperature=2.0)
    system.particles = Particles(dim=3)
    for pos in xyz:
        system.add_particle(pos, vel=np.zeros_like(pos),
                            force=np.zeros_like(pos),
                            mass=1.0, name='Ar', ptype=0)

    gen_settings = {'distribution': 'maxwell', 'momentum': True}
    system.generate_velocities(**gen_settings)
    print(system.particles)

    print('Creating force field:')
    potentials = [PairLennardJonesCutnp(dim=3, shift=True, mixing='geometric')]
    parameters = [{0: {'sigma': 1, 'epsilon': 1, 'rcut': 2.5}}]
    ffield = ForceField('Lennard Jones force field',
                        potential=potentials, params=parameters)
    system.forcefield = ffield
    print(system.forcefield)
    return system


def set_up_simulation(system):
    """Set up the simulation."""
    print('Creating simulation:')
    engine = VelocityVerlet(0.002)
    simulation = SimulationNVE(ensemble={'system': system,
                                         'engine': engine},
                               settings={},
                               controls={'steps': 200})
    return simulation


def run_simulation(simulation):
    """Run the simulation and collect some outputs."""
    ekin = []
    vpot = []
    etot = []
    step = []
    for result in simulation.run():
        if result['cycle']['step'] % 10 == 0:
            print('Step:', result['cycle']['step'])
        step.append(result['cycle']['step'])
        ekin.append(result['thermo']['ekin'])
        vpot.append(result['thermo']['vpot'])
        etot.append(result['thermo']['etot'])
    return step, vpot, ekin, etot


def plot_results(step, vpot, ekin, etot, show=True):
    """Plot some energies."""
    fig1 = plt.figure()
    ax1 = fig1.add_subplot(111)
    ax1.plot(step, ekin, label='Kinetic energy')
    ax1.plot(step, etot, label='Total energy')
    ax1.plot(step, vpot, label='Potential energy')
    ax1.set_xlabel('Step no.')
    ax1.set_ylabel('Energy / reduced units')
    ax1.legend()
    fig1.tight_layout()
    fig1.savefig('out.png', dpi=150)
    if show:
        plt.show()


def main(args):
    """Set up, run simulation and plot some energies."""
    system = set_up_system()
    simulation = set_up_simulation(system)
    step, vpot, ekin, etot = run_simulation(simulation)
    plot_results(step, vpot, ekin, etot, show='noplot' not in args)


if __name__ == '__main__':
    main(sys.argv[1:])



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This file defines the order parameter used for the GROMACS example."""
import os
import logging
from pyretis.orderparameter import Distance
from itertools import combinations
from pyretis.orderparameter import OrderParameter
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


class DistanceL(Distance):
    """Distance(OrderParameter).

    This class computes the distance between the O of water using internal
    function.

    Attributes
    ----------
    name : string
        A human-readable name for the order parameter.

    """

    def __init__(self, idx1, idx2):
        """Set up the order parameter."""
        self.index = (idx1, idx2)
        self.periodic = True



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test using the LAMMPS engine."""
import os
import sys
import colorama
import numpy as np
from matplotlib import pyplot as plt
from pyretis.core.path import Path
from pyretis.inout.common import make_dirs
from pyretis.engines.lammps import LAMMPSEngine
from pyretis.inout import print_to_screen
from pyretis.core.random_gen import create_random_generator
from pyretis.testing.helpers import clean_dir
from pyretis.testing.systemhelp import create_system_ext
from plotting import plot_compare, plot_xy


HERE = os.path.abspath(os.path.dirname(__file__))
STEPS = 100
SUBCYCLES = 2


def run_forward():
    """Use the LAMMPS engine to propagate a MD simulation forward in time."""
    print_to_screen(
        '\nTesting the LAMMPS engine by propagating forward & backward.',
    )
    engine = LAMMPSEngine('lmp_serial', 'lammps_input', SUBCYCLES)
    # Create a dummy system:
    system = create_system_ext(pos=('system.data', 0))
    exe_dir = os.path.join(HERE, 'run-forward-backward')
    make_dirs(exe_dir)
    clean_dir(exe_dir)
    engine.exe_dir = exe_dir
    pathf = Path(rgen=None, maxlen=STEPS)
    print_to_screen('-> Propagating forward....', level='info')
    ensemble = {
        'system': system,
        'order_function': None,
        'interfaces': [-1, 0.0, 2.0]
    }
    engine.propagate(path=pathf, ensemble=ensemble, reverse=False)
    # Propagate from the last point, but backward:
    print_to_screen(
        '-> Setting system to last point in the forward path.',
        level='info'
    )
    last = pathf.phasepoints[-1]
    pathb = Path(rgen=None, maxlen=pathf.length)
    print_to_screen('-> Propagating backward....', level='info')
    ensemble = {
        'system': last,
        'order_function': None,
        'interfaces': [-1, 0.0, 3.0]
    }
    engine.propagate(path=pathb, ensemble=ensemble, reverse=True)
    print_to_screen('-> Comparing forward & backward:', level='info')
    order_f = np.array([i.order for i in pathf.phasepoints])
    order_b = np.array([i.order for i in reversed(pathb.phasepoints)])
    data_sets = [
        [
            (range(len(order_f)), order_f[:, 0], 'Forward.'),
            (range(len(order_b)), order_b[:, 0], 'Backward.'),
        ],
        [
            (range(len(order_f)), order_f[:, 1], 'Forward.'),
            (range(len(order_b)), -1*order_b[:, 1], 'Backward.'),
        ]
    ]
    plot_compare(data_sets, ['OP1', 'OP2'])
    data_sets_xy = [
        (order_f[:, 0], order_b[:, 0], 'OP1, forward', 'OP1, backward'),
        (order_f[:, 1], -1*order_b[:, 1], 'OP2, forward', 'OP2, backward'),
    ]
    plot_xy(data_sets_xy)


def run_backward():
    """Use the LAMMPS engine to propagate a MD simulation backward in time."""
    print_to_screen(
        '\nTesting the LAMMPS engine by propagating backward & forward.',
    )
    engine = LAMMPSEngine('lmp_serial', 'lammps_input', SUBCYCLES)
    # Create a dummy system:
    system = create_system_ext(pos=('system.data', 0))
    exe_dir = os.path.join(HERE, 'run-backward-forward')
    make_dirs(exe_dir)
    clean_dir(exe_dir)
    engine.exe_dir = exe_dir
    rgen = create_random_generator({'seed': 2})
    ensemble = {
        'system': system,
        'rgen': rgen,
    }
    vel_settings = {'aimless': True}
    engine.modify_velocities(ensemble, vel_settings)
    pathb = Path(rgen=None, maxlen=STEPS)
    print_to_screen('-> Propagating backward....', level='info')
    ensemble = {
        'system': system,
        'order_function': None,
        'interfaces': [1.2, 5.0, 12.0],
    }
    engine.propagate(path=pathb, ensemble=ensemble, reverse=True)
    # Propagate from the last point, but forward:
    print_to_screen(
        '-> Setting system to last point in the backward path.',
        level='info'
    )
    last = pathb.phasepoints[-1]
    pathf = Path(rgen=None, maxlen=pathb.length)
    print_to_screen('-> Propagating forward....', level='info')
    ensemble = {
        'system': last,
        'order_function': None,
        'interfaces': [1.0, 5.0, 12.0],
    }
    engine.propagate(path=pathf, ensemble=ensemble, reverse=False)
    print_to_screen('-> Comparing backward & forward:', level='info')
    order_f = np.array([i.order for i in reversed(pathf.phasepoints)])
    order_b = np.array([i.order for i in pathb.phasepoints])
    data_sets = [
        [
            (range(len(order_f)), order_f[:, 0], 'Forward.'),
            (range(len(order_b)), order_b[:, 0], 'Backward.'),
        ],
        [
            (range(len(order_f)), -1*order_f[:, 1], 'Forward.'),
            (range(len(order_b)), order_b[:, 1], 'Backward.'),
        ]
    ]
    plot_compare(data_sets, ['OP1', 'OP2'])
    data_sets_xy = [
        (order_f[:, 0], order_b[:, 0], 'OP1, forward', 'OP1, backward'),
        (-1*order_f[:, 1], order_b[:, 1], 'OP2, forward', 'OP2, backward'),
    ]
    plot_xy(data_sets_xy)


def main(plot=False):
    """Run the comparisons."""
    run_forward()
    run_backward()
    if plot:
        plt.show()


if __name__ == '__main__':
    colorama.init(autoreset=True)
    main(plot=len(sys.argv) >= 2)
